from __future__ import annotations

from typing import ClassVar, Generic, Type, TypedDict, TypeVar


class ElligibleForPlugin:
    """A base class for all classes that should be available to be replaced by plugins."""

    def __new__(cls, *args, **kwargs):
        # hook up plugins here
        for plugin in _active:
            if cls in plugin.override_components:
                return super().__new__(plugin.override_components[cls])

        return super().__new__(cls)


class PluginComponent:
    __plugin__: ClassVar[Plugin]

    def __init_subclass__(cls, **kwargs) -> None:
        cls.__replace_base__ = kwargs.pop("replace_base", False)
        return super().__init_subclass__()


class PluginMeta(type):
    def __new__(mcs, name, bases, namespace, **kwargs):
        # Collect and register plugin components
        _components = []
        _replacements: dict[type, type] = {}

        for attr_name, attr_value in namespace.items():
            if isinstance(attr_value, type) and issubclass(attr_value, PluginComponent):
                _components.append(attr_value)

                # Handle replacement if `replace_base=True` is set
                if getattr(attr_value, "__replace_base__", False):
                    for base in attr_value.__bases__:
                        if issubclass(base, ElligibleForPlugin):
                            _replacements[base] = attr_value
                            break
                    else:
                        raise ValueError(
                            f"Could not find a base class that is elligible for replacement by the plugin API (for {attr_value})"
                        )

        namespace["components"] = _components
        namespace["override_components"] = _replacements

        return super().__new__(mcs, name, bases, namespace)


class BasePluginOpts(TypedDict):
    pass


_T_PluginOpts = TypeVar("_T_PluginOpts", bound=BasePluginOpts)


class Plugin(Generic[_T_PluginOpts], metaclass=PluginMeta):
    name: str
    override_components: dict[type, type]
    components: list[Type[PluginComponent]]

    def __init__(self, opts: _T_PluginOpts) -> None:
        self.opts = opts

        # remember the source plugin for each class
        for cls in self.components:
            cls.__plugin__ = self


# active plugins set
_active: set[Plugin] = set()


def load(*plugins: Plugin) -> None:
    _active.update(plugins)


def unload() -> None:
    _active.clear()



from datetime import datetime
from pathlib import Path
from unittest.mock import patch

import pytest
from sqlalchemy.exc import SQLAlchemyError

from bamboost.index.base import Index
from bamboost.index.store import simulations_upsert_stmt


@pytest.fixture
def index():
    return Index(sql_file=":memory:")


def test_transaction_commit(index: Index):
    with patch.object(index._s, "commit", wraps=index._s.commit) as mock_commit:
        with index.sql_transaction() as session:
            assert session is index._s
        mock_commit.assert_called_once()


def test_transaction_rollback_on_error(index: Index, tmp_path: Path):
    index.upsert_collection("COLL123", tmp_path)

    with patch.object(index._s, "rollback", wraps=index._s.rollback) as mock_rollback:
        with patch.object(index._s, "commit", side_effect=SQLAlchemyError("boom")):
            with pytest.raises(SQLAlchemyError):
                with index.sql_transaction():
                    index._s.execute(
                        simulations_upsert_stmt(
                            {
                                "collection_uid": "COLL123",
                                "name": "apple",
                                "created_at": datetime.now(),
                                "modified_at": datetime.now(),
                                "status": "initialized",
                                "submitted": False,
                            }
                        )
                    )

        # 04.03.25: this setup does not work (?)
        # mock_rollback.assert_called_once()

    # Reset the ORM objects and check that the transaction was rolled back
    index._s.reset()
    assert len(index.all_simulations) == 0



"""
This module provides a high-level abstraction for working with HDF5 (`h5py`) groups and
datasets. It is built on the concept of describing an object in the file with a reference
to it (deterministic reference: a file instance, and a path inside the HDF file).

The reference handles file opening and closing, and provides a simple interface to
data, attributes and subgroups. In essence, this model provides `h5py` objects like any
other in-memory data structure.

Classes:
    H5Reference:
        Base class for Groups and Datasets.
    Group:
        Read-only group reference.
    MutableGroup:
        Mutable group reference.
    Dataset:
        Read-only dataset reference.
"""

from __future__ import annotations

import pkgutil
from enum import IntEnum
from functools import cached_property
from pathlib import Path
from typing import (
    Any,
    Dict,
    Generator,
    Literal,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

import h5py
import numpy as np
from typing_extensions import Self

import bamboost
from bamboost._logger import BAMBOOST_LOGGER
from bamboost._typing import _MT, ArrayLike, Mutable
from bamboost.core.hdf5.attrsdict import AttrsDict
from bamboost.core.hdf5.file import (
    FileMode,
    H5Object,
    HDF5File,
    mutable_only,
    with_file_open,
)
from bamboost.core.hdf5.filemap import FilteredFileMap
from bamboost.core.hdf5.hdf5path import HDF5Path
from bamboost.mpi import MPI_ON

log = BAMBOOST_LOGGER.getChild("hdf5")

MPI_ACTIVE = "mpio" in h5py.registered_drivers() and h5py.get_config().mpi and MPI_ON

_RT_group = TypeVar("_RT_group", bound=Union["Group", "Dataset"])


class InvalidReferenceError(KeyError):
    def __init__(self, path: HDF5Path, filename: str):
        super().__init__(f"Object {path} not found in file {filename}")


class RefStatus(IntEnum):
    INVALID = 0
    VALID = 1
    NOT_CHECKED = 2


class H5Reference(H5Object[_MT]):
    _status: RefStatus = RefStatus.NOT_CHECKED

    def __init__(self, path: str, file: HDF5File[_MT]):
        self._file = file
        self._path = HDF5Path(path)

        # We validate the existance of the object in the file by first, checking the file
        # map (avoiding a file operation), and if it is not there, we check the file.
        if self._path in self._file.file_map:
            self._status = RefStatus.VALID
            return

        # If we deal with an immutable file, we can't create a new reference.
        # We decide to immediately check the validity of the reference and raise an
        # Exception if it is invalid.
        if not self._file.mutable:
            if not self._is_valid():
                raise InvalidReferenceError(self._path, self._file._filename)
            self._status = RefStatus.VALID

    @property
    def _obj(self) -> Union[h5py.Group, h5py.Dataset, h5py.Datatype]:
        """Returns the h5py object bound at the path of this reference.

        Raises:
            KeyError: 'Unable to synchronously open object (invalid identifier type to function)'
                If the file is not open. Or if the object at the path does not exist.
        """
        _obj = self._file[self._path]
        self._status = RefStatus.VALID
        return _obj

    @with_file_open(FileMode.READ)
    def _is_valid(self):
        try:
            _ = self._obj
            return True
        except KeyError:
            return False

    def __repr__(self) -> str:
        valid_str = (
            "NOT_CHECKED"
            if self._status == RefStatus.NOT_CHECKED
            else "VALID"
            if self._status == RefStatus.VALID
            else "INVALID"
        )
        return f'<HDF5 {type(self).__name__} "{self._path}" ({valid_str}, file {self._file._filename})>'

    @overload
    def __getitem__(self, value: str) -> Union[Self, Dataset]: ...
    @overload
    def __getitem__(self, value: Union[slice, Tuple[()]]) -> Any: ...
    @overload
    def __getitem__(self, value: tuple[str, Type[Group]]) -> Group[_MT]: ...
    @overload
    def __getitem__(self, value: tuple[str, Type[Dataset]]) -> Dataset[_MT]: ...
    def __getitem__(self, value):
        if isinstance(value, str):
            if _type := self._file.file_map.get(self._path / value):
                return self.new(
                    self._path.joinpath(value),
                    self._file,
                    Group if _type == h5py.Group else Dataset,
                )

        with self._file.open(FileMode.READ):
            obj = self._obj
            assert not isinstance(obj, h5py.Datatype), (
                "__getitem__ not implemented for Datatype"
            )

            # If the value is a slice or empty tuple, we return the sliced dataset
            if isinstance(value, slice) or (
                isinstance(value, tuple) and len(value) == 0
            ):
                return obj[value]

            # Here we know that we are looking for a group or dataset
            if isinstance(value, tuple):
                name, _type = value
            else:
                name, _type = cast(str, value), None
            return self.new(self._path / name, self._file, _type)

    @classmethod
    def new(
        cls,
        path: str,
        file: HDF5File[_MT],
        _type: Optional[Type[_RT_group]] = None,
    ) -> _RT_group:
        """Returns a new pointer object."""
        if _type:
            return _type(path, file)

        with file.open(FileMode.READ):
            _obj = file[path]
            if isinstance(_obj, h5py.Group):
                return cast(_RT_group, Group(path, file))
            elif isinstance(_obj, h5py.Dataset):
                return cast(_RT_group, Dataset(path, file))
            else:
                raise ValueError(f"Object {path} is not a group or dataset")

    @cached_property
    def attrs(self) -> AttrsDict[_MT]:
        return AttrsDict(self._file, self._path)

    @property
    @with_file_open(FileMode.READ)
    def parent(self) -> Group[_MT]:
        return Group(self._obj.parent.name or "", self._file)


class Group(H5Reference[_MT]):
    def __init__(self, path, file):
        super().__init__(path, file)

        # Create a subset view of the file map with all objects
        self._group_map = FilteredFileMap(file.file_map, path)

        # In addition to the base class, we test if the object is not a group
        # first check by file map, fallback: check with file operation
        # Try except for edge case where the file is not created yet.
        try:
            _valid = (file.file_map.get(self._path) is h5py.Group) or self._is_valid()
        except FileNotFoundError:
            _valid = True
        self._status = RefStatus(_valid)

    @mutable_only
    def __setitem__(self: Group[Mutable], key, newvalue):
        """Used to set an attribute.
        Will be written as an attribute to the group.
        """
        if isinstance(newvalue, np.ndarray):
            self.add_numerical_dataset(key, np.array(newvalue))
        else:
            self.attrs.__setitem__(key, newvalue)

    @mutable_only
    @with_file_open(FileMode.APPEND)
    def __delitem__(self: Group[Mutable], key) -> None:
        """Deletes an item."""
        if key in self.attrs.keys():
            del self.attrs[key]
        else:
            self._file.delete_object(self._path / key)

    def __contains__(self, key: str) -> bool:
        return key in self._group_map

    def _ipython_key_completions_(self):
        return self.keys()

    @property
    def _obj(self) -> h5py.Group:
        _obj = super()._obj
        if not isinstance(_obj, h5py.Group):
            raise ValueError(f"Object {self._path} is not a group")
        return _obj

    @with_file_open(FileMode.READ)
    def _is_valid(self):
        try:
            obj = self._obj
            return isinstance(obj, h5py.Group)
        except KeyError:
            return False

    def __iter__(self):
        for key in self.keys():
            yield self.__getitem__(key)

    def _assert_file_map_is_valid(self):
        if not self._group_map.valid:
            with self._file.open(FileMode.READ):
                self._group_map.file_map.populate()

    def keys(self):
        self._assert_file_map_is_valid()
        return self._group_map.keys()

    def groups(self):
        self._assert_file_map_is_valid()
        return self._group_map.children_groups()

    def datasets(self):
        self._assert_file_map_is_valid()
        return self._group_map.children_datasets()

    def items(
        self,
        *,
        filter: Optional[Literal["groups", "datasets"]] = None,
    ) -> Generator[Tuple[str, Union[Group[_MT], Dataset[_MT]]], None, None]:
        if filter:
            keys = (
                self._group_map.children_groups()
                if filter == "groups"
                else self._group_map.children_datasets()
            )
        else:
            keys = self._group_map.children()

        for key in keys:
            yield key, self.__getitem__(key)

    @with_file_open(FileMode.READ)
    def _repr_html_(self):
        """Repr showing the content of the group."""
        # If the object is not a group, return a simple representation
        try:
            _obj = self._obj
        except ValueError:
            self._status = RefStatus.INVALID
            return f"Invalid HDF5 object: <b>{self._path}</b> is not a group"
        except KeyError:
            self._status = RefStatus.INVALID
            return f"Invalid HDF5 object: <b>{self._path}</b> not found in file"

        from jinja2 import Template

        attrs = self.attrs
        groups = {key: len(_obj[self._path / key]) for key in self.groups()}  # type: ignore
        datasets = {
            key: (_obj[self._path / key].dtype, _obj[self._path / key].shape)  # type: ignore
            for key in self.datasets()
        }

        path = self._path
        path = path if path[0] == "/" else "/" + path

        html_template = pkgutil.get_data(
            bamboost.__name__, "_repr/hdf5_group.html"
        ).decode()  # type: ignore
        icon = pkgutil.get_data(bamboost.__name__, "_repr/icon.txt").decode()  # type: ignore
        template = Template(html_template)

        return template.render(
            uid=Path(self._file._filename).parent.name,
            name=path,
            icon=icon,
            version=bamboost.__version__,
            attrs=attrs,
            groups=groups,
            datasets=datasets,
        )

    @mutable_only
    @with_file_open(FileMode.APPEND)
    def require_self(self: Group[Mutable]) -> None:
        """Create the group if it doesn't exist yet."""
        self._file.require_group(self._path)
        self._file.file_map[self._path] = h5py.Group
        self._status = RefStatus.VALID

    @overload
    def require_group(
        self: Group[Mutable],
        name: str,
        *,
        return_type: Type[_RT_group],
    ) -> _RT_group: ...
    @overload
    def require_group(self: Group[Mutable], name: str) -> Group[Mutable]: ...
    @mutable_only
    @with_file_open(FileMode.APPEND, driver="mpio")
    def require_group(self, name, *, return_type=None):
        """Create a group if it doesn't exist yet."""
        self._obj.require_group(name)

        # update file_map
        self._group_map[name] = h5py.Group

        return self.new(self._path.joinpath(name), self._file, _type=return_type)

    @mutable_only
    def require_dataset(
        self: Group[Mutable],
        name: str,
        shape: tuple[int, ...],
        dtype,
        exact: bool = False,
        **kwargs,
    ) -> h5py.Dataset:
        grp = self._obj.require_dataset(name, shape, dtype, exact=exact, **kwargs)
        self._group_map[name] = h5py.Dataset  # type: ignore
        return grp

    @mutable_only
    def add_numerical_dataset(
        self: Group[Mutable],
        name: str,
        vector: ArrayLike,
        attrs: Optional[Dict[str, Any]] = None,
        dtype: Optional[str] = None,
        *,
        file_map: bool = True,
    ) -> None:
        """Add a dataset to the group. Error is thrown if attempting to overwrite
        with different shape than before. If same shape, data is overwritten
        (this is inherited from h5py -> require_dataset)

        Args:
            name: Name for the dataset
            vector: Data to write (max 2d)
            attrs: Optional. Attributes of dataset.
            dtype: Optional. dtype of dataset. If not specified, uses dtype of inpyt array
            file_map: Optional. If True, the dataset is added to the file map. Default is True.
        """
        if attrs is None:
            attrs = {}
        length_local = vector.shape[0]
        length_p = np.array(self._file._comm.allgather(length_local))
        length = np.sum(length_p)
        dim = vector.shape[1:]
        vec_shape = length, *dim

        ranks = np.array([i for i in range(self._file._comm.size)])
        idx_start = np.sum(length_p[ranks < self._file._comm.rank])
        idx_end = idx_start + length_local

        with self._file.open(FileMode.APPEND, driver="mpio"):
            dataset = self._obj.require_dataset(
                name, shape=vec_shape, dtype=dtype if dtype else vector.dtype
            )
            dataset[idx_start:idx_end] = vector

            def _write_attrs():
                self._obj[name].attrs.update(attrs)

            self.post_write_instruction(_write_attrs)

        log.info(f'Written dataset to "{self._path}/{name}"')

        # update file_map
        if file_map:
            self._group_map[name] = h5py.Dataset

    @mutable_only
    def add_dataset(
        self: Group[Mutable],
        name: str,
        data: Any,
        attrs: Optional[Dict[str, Any]] = None,
        dtype: Optional[str] = None,
    ) -> None:
        def _write_instruction():
            self._obj.create_dataset(name, data=data, dtype=dtype)
            if attrs:
                self._obj[name].attrs.update(attrs)
            log.info(f'Written dataset to "{self._path}/{name}"')

        self.post_write_instruction(_write_instruction)

        # update file_map
        self._group_map[name] = h5py.Dataset


class Dataset(H5Reference[_MT]):
    @property
    def _obj(self) -> h5py.Dataset:
        obj = super()._obj
        if not isinstance(obj, h5py.Dataset):
            raise ValueError(f"Object {self._path} is not a dataset")
        return obj

    @with_file_open(FileMode.READ)
    def __getitem__(self, key: tuple | slice | int) -> Any:
        return h5py.Dataset.__getitem__(self._obj, key)

    @property
    @with_file_open(FileMode.READ)
    def shape(self) -> tuple[int, ...]:
        return self._obj.shape

    @property
    @with_file_open(FileMode.READ)
    def dtype(self):
        return self._obj.dtype

    @property
    @with_file_open(FileMode.READ)
    def array(self) -> np.ndarray:
        return self._obj[()]


# Aliases for Group and Dataset
_g = Group
_d = Dataset



import os
from datetime import datetime
from pathlib import Path
from typing import Type
from unittest.mock import ANY, MagicMock, patch

import pytest

from bamboost import constants
from bamboost.core.collection import Collection
from bamboost.core.hdf5.attrsdict import AttrsDict
from bamboost.core.hdf5.ref import Group
from bamboost.core.simulation.base import (
    Simulation,
    SimulationWriter,
    Status,
    StatusInfo,
    _Simulation,
)
from bamboost.core.simulation.series import Series
from bamboost.index import SimulationName, SimulationUID


@pytest.fixture
def tmp_simulation(tmp_path: Path) -> Simulation:
    # create collection
    Collection(tmp_path)

    # create a directory for the simulation (simulates coll.add)
    tmp_path.joinpath("test").mkdir()
    sim = SimulationWriter("test", tmp_path)
    # create the file
    sim.initialize()

    return Simulation("test", tmp_path)


@pytest.fixture
def tmp_simulation_writer(tmp_path: Path) -> SimulationWriter:
    # create collection
    Collection(tmp_path)

    # create a directory for the simulation (simulates coll.add)
    tmp_path.joinpath("test").mkdir()
    sim = SimulationWriter("test", tmp_path)
    # create the file
    sim.initialize()

    return sim


@pytest.fixture
def mock_hdf5():
    """Mock the HDF5 file interactions."""
    with patch("bamboost.core.hdf5.file.HDF5File", autospec=True) as mock_hdf5:
        yield mock_hdf5


def test_import():
    assert Simulation
    assert SimulationWriter


def test_update_database(tmp_collection: Collection):
    sim = tmp_collection.add("test_update_database")
    index = tmp_collection._index

    new_metadata = {
        "ignored": "ignored",
        "status": "test",
        "description": "test",
        "tags": ["first", "first"],
    }
    sim.update_database(metadata=new_metadata)
    uid = tmp_collection.uid
    assert index.simulation(SimulationUID(uid, sim.name)).status == "test"
    assert index.simulation(SimulationUID(uid, sim.name)).description == "test"
    assert index.simulation(SimulationUID(uid, sim.name)).tags == ["first"]

    new_parameters = {"last_name": "Bourne", "first_name": "Jane"}
    sim.update_database(parameters=new_parameters)
    assert (
        index.simulation(SimulationUID(uid, sim.name)).parameter_dict == new_parameters
    )


@pytest.mark.parametrize(
    "status_string, expected",
    [
        ("initialized", Status.INITIALIZED),
        ("started", Status.STARTED),
        ("finished", Status.FINISHED),
        ("failed", Status.FAILED),
        ("some random stuff", Status.UNKNOWN),
        ("", Status.UNKNOWN),
    ],
)
def test_status_info_parse(status_string, expected):
    status_info = StatusInfo.parse(status_string)

    assert status_info.status == expected


def test_status_info_message():
    status_string = "failed [RuntimeError - 'Some error']"
    status_info = StatusInfo.parse(status_string)
    assert status_info.status == Status.FAILED
    assert status_info.message == "RuntimeError - 'Some error'"


def test_simulation_name_given():
    name = SimulationName("test")
    assert name == "test"


def test_simulation_name_generation():
    length = 10
    name = SimulationName(length=length)
    assert name is not None
    assert len(name) == length


def test_init(tmp_collection: Collection):
    params = {"test": "test"}
    tmp_collection.add("test", parameters=params)

    sim = Simulation("test", tmp_collection.path)
    assert sim.name == "test"
    assert sim.path == tmp_collection.path.joinpath("test")
    assert sim._index  # reference to index database exists
    assert sim.collection_uid == tmp_collection.uid


def test_fail_non_existing(tmp_collection_burn: Collection):
    """Test that init a non-existing simulation raises FileNotFoundError"""
    with pytest.raises(FileNotFoundError):
        Simulation("test", tmp_collection_burn.path)


def test_html_repr(tmp_collection_burn: Collection):
    sim = tmp_collection_burn.add("test")
    assert sim._repr_html_() is not None


def test_root_group(test_collection: Collection):
    sim = test_collection["testsim1"]
    assert isinstance(sim.root, Group)


@pytest.mark.parametrize(
    "cls, mutable", [(Simulation, False), (SimulationWriter, True)]
)
def test_mutability(test_collection: Collection, cls: Type[_Simulation], mutable: bool):
    path = test_collection.path
    sim = cls("testsim1", path)
    assert sim.mutable == mutable


def test_from_uid(test_collection: Collection):
    uid = test_collection.uid
    sim_uid = f"{uid}{constants.UID_SEPARATOR}testsim1"
    sim = Simulation.from_uid(sim_uid)
    sim2 = Simulation("testsim1", test_collection.path)
    assert sim == sim2


def test_get_writer_from_simulation(test_collection: Collection):
    sim = test_collection["testsim1"]
    writer = sim.edit()
    assert isinstance(writer, SimulationWriter)


@pytest.mark.skip("Dumb test")
def test_parameters(test_collection: Collection):
    sim = test_collection["testsim1"]
    assert isinstance(sim.parameters, AttrsDict)


def test_status(test_collection: Collection):
    sim = test_collection["testsim1"]
    assert sim.status == StatusInfo(Status.INITIALIZED)


def test_created_at_is_datetime(test_collection: Collection):
    sim = test_collection["testsim1"]
    assert isinstance(sim.created_at, datetime)


def test_simulation_enter_path(tmp_simulation: Simulation):
    initial_dir = Path.cwd()

    with tmp_simulation.enter_path():
        assert Path.cwd() == tmp_simulation.path

    assert Path.cwd() == initial_dir


def test_create_xdmf(tmp_simulation):
    """Test that create_xdmf correctly calls XDMFWriter and writes an XDMF file."""
    tmp_simulation = tmp_simulation.edit()

    with (
        patch("bamboost.core.simulation.xdmf.XDMFWriter") as mock_xdmf_writer,
        patch.object(tmp_simulation, "meshes") as mock_group_meshes,
    ):
        mock_group_meshes.__getitem__.return_value = "test_mesh"
        mock_writer_instance = mock_xdmf_writer.return_value
        mock_writer_instance.write_file = MagicMock()

        field_names = ["velocity", "pressure"]
        timesteps = [0.0, 0.1, 0.2]
        filename = tmp_simulation.path / "test_output.xdmf"
        mesh_name = "test_mesh"

        tmp_simulation.create_xdmf(
            field_names=field_names,
            timesteps=timesteps,
            filename=filename,
            mesh_name=mesh_name,
        )

        # Ensure XDMFWriter was instantiated with the correct file
        mock_xdmf_writer.assert_called_once_with(tmp_simulation._file)

        # Ensure methods were called with expected arguments
        mock_writer_instance.add_mesh.assert_called_once_with("test_mesh")
        mock_writer_instance.add_timeseries.assert_called_once_with(
            timesteps, tmp_simulation.data.get_fields(*field_names), mesh_name
        )
        mock_writer_instance.write_file.assert_called_once_with(filename)


# SimulationWriter tests
def test_writer_context_success(tmp_simulation: Simulation):
    writer = tmp_simulation.edit()

    with writer as w:
        assert w.status.status == Status.STARTED

    assert writer.status.status == Status.FINISHED


def test_writer_context_failure(tmp_simulation: Simulation):
    writer = tmp_simulation.edit()

    with pytest.raises(ValueError):
        with writer as w:
            assert w.status.status == Status.STARTED
            raise ValueError("test")

    assert writer.status.status == Status.FAILED


def test_simulation_writer_set_status(tmp_simulation_writer):
    """Test that setting the status updates metadata correctly."""
    tmp_simulation_writer.status = StatusInfo(Status.STARTED)
    assert tmp_simulation_writer.metadata["status"] == "started"

    tmp_simulation_writer.status = StatusInfo(Status.FAILED, "An error occurred")
    assert tmp_simulation_writer.metadata["status"] == "failed [An error occurred]"


def test_simulation_writer_set_description(tmp_simulation_writer):
    """Test that setting the description updates metadata correctly."""
    tmp_simulation_writer.description = "This is a test description"
    assert tmp_simulation_writer.metadata["description"] == "This is a test description"


def test_simulation_writer_set_tags(tmp_simulation_writer):
    """Test that setting tags updates metadata correctly."""
    tmp_simulation_writer.tags = ["foo", "bar", "foo", ""]
    # the tags are always stored as a set!
    # so the returned tags are of type set and duplicates and empty strings are removed
    assert tmp_simulation_writer.metadata["tags"] == {"foo", "bar"}
    assert tmp_simulation_writer.tags == {"foo", "bar"}


def test_simulation_writer_require_series(tmp_simulation_writer):
    """Test that require_series initializes the series in the HDF5 file if missing."""
    with patch.object(
        tmp_simulation_writer,
        "_initialize_series",
        wraps=tmp_simulation_writer._initialize_series,
    ) as mock_init_series:
        series = tmp_simulation_writer.require_series("custom_series")

        # Ensure _initialize_series is called since the series was missing
        mock_init_series.assert_called_once_with("custom_series")

        # Ensure the correct object is returned
        assert isinstance(series, Series)


def test_simulation_writer_copy_files(tmp_simulation_writer, tmp_path):
    """Test that copy_files correctly copies files and directories to the simulation folder."""
    source_file = tmp_path / "test_file.txt"
    source_file.write_text("Test content")

    source_dir = tmp_path / "test_dir"
    source_dir.mkdir()
    (source_dir / "file_in_dir.txt").write_text("Inside directory")

    with patch("shutil.copy") as mock_copy, patch("shutil.copytree") as mock_copytree:
        tmp_simulation_writer.copy_files([source_file, source_dir])

        # Ensure shutil.copy was called for the file
        mock_copy.assert_called_once_with(source_file, tmp_simulation_writer.path)

        # Ensure shutil.copytree was called for the directory
        mock_copytree.assert_called_once_with(source_dir, tmp_simulation_writer.path)


def test_simulation_writer_create_run_script(tmp_simulation_writer):
    """Test that create_run_script correctly generates a bash script."""
    script_path = tmp_simulation_writer.path / constants.RUN_FILE_NAME
    commands = ["echo 'Hello'", "echo 'World'"]
    sbatch_kwargs = {"--time": "02:00:00", "--ntasks": "4"}

    # Test with euler=True (should include sbatch options)
    tmp_simulation_writer.create_run_script(
        commands, euler=True, sbatch_kwargs=sbatch_kwargs
    )

    assert script_path.exists()

    with script_path.open() as f:
        content = f.read()

    # Ensure sbatch options are included
    assert "#SBATCH --time=02:00:00" in content
    assert "#SBATCH --ntasks=4" in content
    assert (
        f"#SBATCH --output={tmp_simulation_writer.path.joinpath(tmp_simulation_writer.name + '.out')}"
        in content
    )
    assert f"#SBATCH --job-name={tmp_simulation_writer.uid}" in content

    # Ensure commands are included
    assert "echo 'Hello'" in content
    assert "echo 'World'" in content

    # Ensure environment variables are set
    assert f"SIMULATION_DIR={tmp_simulation_writer.path.as_posix()}" in content
    assert f"SIMULATION_ID={tmp_simulation_writer.uid}" in content

    # Ensure metadata is updated
    assert tmp_simulation_writer.metadata["submitted"] is False

    # Test with euler=False (should NOT include sbatch options)
    tmp_simulation_writer.create_run_script(commands, euler=False)

    with script_path.open() as f:
        content_no_sbatch = f.read()

    assert "#SBATCH" not in content_no_sbatch  # No sbatch directives
    assert "echo 'Hello'" in content_no_sbatch
    assert "echo 'World'" in content_no_sbatch


def test_simulation_writer_run_simulation(tmp_simulation_writer: SimulationWriter):
    """Test that run_simulation correctly executes the run script."""
    # Ensure the script exists
    script_path = tmp_simulation_writer._bash_file
    script_path.touch()  # Create an empty file

    with (
        patch("subprocess.run") as mock_run,
        patch.dict(os.environ, {"BAMBOOST_MPI": "1"}, clear=False),
    ):
        tmp_simulation_writer.run_simulation(executable="bash")

        # Ensure subprocess.run is called with the correct arguments
        mock_run.assert_called_once_with(["bash", script_path.as_posix()], env=ANY)

        # Ensure BAMBOOST_MPI is removed from environment
        _, kwargs = mock_run.call_args
        assert "BAMBOOST_MPI" not in kwargs["env"]

        # Ensure metadata is updated
        assert tmp_simulation_writer.metadata["submitted"] is True


def test_simulation_writer_run_simulation_missing_script(tmp_simulation_writer):
    """Test that run_simulation raises an error if the script is missing."""
    with pytest.raises(FileNotFoundError, match="Run script .* does not exist"):
        tmp_simulation_writer.run_simulation(executable="bash")


def test_simulation_writer_submit_simulation(tmp_simulation_writer):
    """Test that submit_simulation correctly calls sbatch."""
    with patch.object(tmp_simulation_writer, "run_simulation") as mock_run_simulation:
        tmp_simulation_writer.submit_simulation()

        # Ensure run_simulation is called with sbatch
        mock_run_simulation.assert_called_once_with(executable="sbatch")



"""Module for indexing BAMBOOST collections.

Uses a caching mechanism using SQLAlchemy and an SQLite database to store
information about collections and simulations.

Usage:
    Create an instance of the `Index` class and use its methods to interact
    with the index.

    >>> from bamboost.index import Index
    >>> index = Index()

    Scan for collections in known paths:

    >>> index.scan_for_collections()

    Resolve the path of a collection:

    >>> index.resolve_path(<collection-uid>)

    Get a simulation from its collection and simulation name:

    >>> index.get_simulation(<collection-uid>, <simulation-name>)

Classes:
    Index: API for indexing BAMBOOST collections and simulations.
"""

from .base import Index
from .scanner import (
    create_identifier_file,
    get_identifier_filename,
)
from .uids import (
    CollectionUID,
    SimulationName,
    SimulationUID,
)

__all__ = [
    "CollectionUID",
    "SimulationUID",
    "SimulationName",
    "Index",
    "create_identifier_file",
    "get_identifier_filename",
]



from typing import Annotated

import typer

from bamboost.cli import _completion
from bamboost.cli.common import console

subapp_alias = typer.Typer(
    name="alias", help="Manage collection aliases in the index.", no_args_is_help=True
)


@subapp_alias.command()
def add(
    uid: str = typer.Argument(
        ...,
        autocompletion=_completion._get_uids_from_db,
        help="The unique ID of the collection to add an alias for.",
    ),
    alias: str = typer.Argument(..., help="The alias to add to the collection."),
) -> None:
    """Add an alias to a collection in the index by its unique ID."""
    with console.status(
        f"[bold blue]Adding alias '{alias}' to collection '{uid}'...", spinner="dots"
    ) as status:
        try:
            from bamboost.core.collection import Collection

            coll = Collection(uid=uid)
            meta = coll.metadata
            if alias in meta.aliases:
                raise ValueError(
                    f"Alias '{alias}' already exists in collection '{uid}'."
                )

            meta.aliases.append(alias)
            meta.save()
            console.print(
                f"[green]:heavy_check_mark: Alias '{alias}' added to collection '{uid}'."
            )
        except Exception as e:
            console.print(f"[bold red]Task failed: {e}")


@subapp_alias.command()
def remove(
    uid: str = typer.Argument(
        ...,
        autocompletion=_completion._get_uids_from_db,
        help="The unique ID of the collection to remove an alias from.",
    ),
    alias: str = typer.Argument(
        ...,
        help="The alias to remove from the collection.",
        autocompletion=_completion._get_aliases_of_collection,
    ),
) -> None:
    """Remove an alias from a collection in the index by its unique ID."""
    with console.status(
        f"[bold blue]Removing alias '{alias}' from collection '{uid}'...",
        spinner="dots",
    ) as status:
        try:
            from bamboost.core.collection import Collection

            coll = Collection(uid=uid)
            meta = coll.metadata
            if alias not in meta.aliases:
                raise ValueError(f"Alias '{alias}' not found in collection '{uid}'.")

            meta.aliases.remove(alias)
            meta.save()
            console.print(
                f"[green]:heavy_check_mark: Alias '{alias}' removed from collection '{uid}'."
            )
        except Exception as e:
            console.print(f"[bold red]Task failed: {e}")


@subapp_alias.command()
def get(
    alias: Annotated[str, typer.Argument(..., help="The alias to look up.")],
) -> None:
    """Get the unique ID of a collection by its alias."""
    with console.status(
        f"[bold blue]Looking up collection by alias '{alias}'...", spinner="dots"
    ) as status:
        from bamboost.index import Index

        ind = Index.default
        with ind.sql_transaction():
            uid = ind._find_collection_uid_by_alias(alias)
            if uid is None:
                raise ValueError(f"Alias '{alias}' not found in any collection.")
        print(uid)



from pathlib import Path, PurePosixPath
from unittest.mock import patch

import h5py
import pytest

from bamboost._typing import Mutable
from bamboost.core.hdf5 import HDF5File, HDF5Path
from bamboost.core.hdf5.filemap import FileMap, FilteredFileMap


@pytest.fixture
def hdf5_file(tmp_path: Path):
    return HDF5File(tmp_path.joinpath("test.h5"), mutable=True)


@pytest.fixture(scope="module")
def hdf5_file_module(tmp_path_module: Path):
    return HDF5File(tmp_path_module.joinpath("test.h5"), mutable=True)


# ------------------------------
# HDF5Path
# ------------------------------
def test_hdf5_path():
    assert HDF5Path("path/to/file", absolute=True) == "/path/to/file"
    assert HDF5Path("//////path////to//file", absolute=True) == "/path/to/file"
    assert HDF5Path("path/to/file", absolute=False) == "path/to/file"
    assert HDF5Path("/path/to/file", absolute=False) == "path/to/file"


def test_hdf5_path_relative_to():
    path = HDF5Path("/path/to/file")
    assert path.relative_to("path") == "to/file"
    assert path.relative_to("path/to") == "file"


def test_hdf5_path_relative_to_error():
    path = HDF5Path("/path/to/file")
    with pytest.raises(ValueError):
        path.relative_to("not_a_parent")


def test_hdf5_path_join():
    path = HDF5Path("/path/to/file")
    assert path.joinpath("another") == "/path/to/file/another"
    # also testing the __truediv__ operator
    assert path / "another" == "/path/to/file/another"


def test_hdf5_path_parent():
    path = HDF5Path("/path/to/file")
    assert path.parent == "/path/to"


def test_hdf5_path_basename():
    path = HDF5Path("/path/to/file")
    assert path.basename == "file"


def test_hdf5_path_posix():
    path = HDF5Path("/path/to/file")
    assert isinstance(path.path, PurePosixPath)
    assert path.path.stem == "file"


# ------------------------------
# FileMap
# ------------------------------
@pytest.fixture(scope="module")
def filemap(hdf5_file_module: HDF5File):
    # add some groups and datasets using h5py directly
    # lenght: 5
    with hdf5_file_module.open("w") as f:
        f.create_group("group1")
        grp2 = f.create_group("group2")
        f.create_dataset("dataset1", data=[1, 2, 3])
        grp2.create_dataset("dataset2", data=[4, 5, 6])
        grp2.create_group("group3")

    # check that the file map is populated
    fm = FileMap(hdf5_file_module)
    with hdf5_file_module.open("r"):
        fm.populate()
    return fm


def test_filemap_populate(hdf5_file: HDF5File):
    # add some groups and datasets using h5py directly
    with hdf5_file.open("w") as f:
        f.create_group("group1")
        f.create_dataset("dataset1", data=[1, 2, 3])

    # check that the file map is populated
    fm = FileMap(hdf5_file)
    with hdf5_file.open("r"):
        with patch("h5py.File.visititems", wraps=fm._file.visititems) as visititems:
            fm.populate()

    # check that the visititems method was called
    visititems.assert_called_once()
    # assert dict is not empty
    assert fm._dict


def test_filemap_populate_error(hdf5_file: HDF5File):
    # check that the method raises an error if the file is not open or was never opened
    with pytest.raises((ValueError, AttributeError)):
        FileMap(hdf5_file).populate()


def test_filemap_getitem(filemap: FileMap):
    # check the entries
    assert filemap["/group1"] == h5py.Group
    assert filemap["/group2"] == h5py.Group
    assert filemap["/group2/dataset2"] == h5py.Dataset
    assert filemap["dataset1"] == h5py.Dataset


def test_filemap_len(filemap: FileMap):
    assert len(filemap) == 5


def test_filemap_iter(filemap: FileMap):
    for key in filemap:
        assert key in [
            "/group1",
            "/group2",
            "/group2/dataset2",
            "/group2/group3",
            "/dataset1",
        ]


def test_filemap_datasets(filemap: FileMap):
    datasets = filemap.datasets()
    assert len(datasets) == 2
    assert "/dataset1" in datasets
    assert "/group2/dataset2" in datasets


def test_filemap_groups(filemap: FileMap):
    groups = filemap.groups()
    assert len(groups) == 3
    assert "/group1" in groups
    assert "/group2" in groups
    assert "/group2/group3" in groups


def test_filtered_filemap(filemap: FileMap):
    filtered = FilteredFileMap(filemap, parent="group2")

    assert len(filtered) == 2
    assert filtered.datasets() == ("dataset2",)
    assert filtered.groups() == ("group3",)
    assert filtered["dataset2"] == h5py.Dataset
    assert filtered["group3"] == h5py.Group


# ------------------------------
# HDF5File
# ------------------------------
@pytest.mark.parametrize(
    "name, mutable",
    [
        ("test.h5", True),
        ("....++.h5", True),  # aweful names are not forbidden
    ],
)
def test_hdf5_file_init(tmp_path: Path, name: str, mutable: bool):
    f = HDF5File(tmp_path.joinpath(name), mutable=mutable)

    # check the attributes
    assert f._filename == tmp_path.joinpath(name).as_posix()
    assert f._path == tmp_path.joinpath(name).absolute()
    assert isinstance(f.file_map, FileMap)
    assert f.mutable == mutable


def test_hdf5_file_init_error(tmp_path: Path):
    """Test that we immediately raise an exception if the file is immutable and does not exist."""
    with pytest.raises(FileNotFoundError):
        HDF5File(tmp_path.joinpath("not_existing.h5"), mutable=False)


def test_hdf5_file_open(hdf5_file: HDF5File):
    with hdf5_file.open("w") as f:
        assert f.file.mode == "r+"
        assert f.file.id.valid


def test_hdf5_file_mutability(hdf5_file: HDF5File[Mutable]):
    hdf5_file._create_file()

    # create an immutable file object and try to open it
    f_immutable = HDF5File(hdf5_file._path, mutable=False)

    with pytest.raises(PermissionError):
        f_immutable.open("w")


def test_hdf5_file_context(hdf5_file: HDF5File):
    with patch("h5py.File.__init__", wraps=h5py.File.__init__) as h5py_init:
        with hdf5_file.open("w"):
            with hdf5_file.open("w"):
                with hdf5_file.open("w"):
                    assert hdf5_file.is_open
    # check that the file was opened only once
    h5py_init.assert_called_once()


def test_hdf5_file_context_mode_change(hdf5_file: HDF5File):
    with hdf5_file.open("w") as f:
        pass  # create file
    with (
        patch("h5py.File.__init__", wraps=h5py.File.__init__) as h5py_init,
        patch("h5py.File.close", wraps=super(HDF5File, hdf5_file).close) as h5py_close,
    ):
        with hdf5_file.open("r"):
            assert hdf5_file.mode == "r"
            with hdf5_file.open("w"):
                assert hdf5_file.mode == "r+"
                with hdf5_file.open("r"):
                    assert hdf5_file.mode == "r+"  # still writable

    # check that the file was opened twice
    assert h5py_init.call_count == 2
    assert h5py_close.call_count == 2


def test_hdf5_file_close_depending_on_context(hdf5_file: HDF5File):
    hdf5_file.open("w")
    hdf5_file.open("w")
    hdf5_file.open("w")
    hdf5_file.close()
    # file should be still open with context stack 2
    assert hdf5_file.is_open
    assert hdf5_file._context_stack == 2

    # close 2 times more to close
    hdf5_file.close()
    hdf5_file.close()
    assert not hdf5_file.is_open
    assert hdf5_file._context_stack == 0

    # check that further calls to close do nothing
    hdf5_file.close()


def test_hdf5_file_force_close(hdf5_file: HDF5File):
    hdf5_file.open("w")
    hdf5_file.open("w")
    hdf5_file.open("w")
    hdf5_file.force_close()
    assert not hdf5_file.is_open



import importlib
import os

import pytest
from _pytest.monkeypatch import MonkeyPatch

from bamboost.mpi import (
    _assert_h5py_has_mpi_support,
    _detect_if_mpi_needed,
    get_mpi_from_env,
)


@pytest.fixture
def h5py_mpi_enabled(monkeypatch):
    class FakeConfig:
        mpi = True

    monkeypatch.setattr("h5py.get_config", lambda: FakeConfig())
    yield


@pytest.fixture
def h5py_mpi_disabled(monkeypatch):
    class FakeConfig:
        mpi = False

    monkeypatch.setattr("h5py.get_config", lambda: FakeConfig())
    yield


def test_detect_mpi_disabled_by_config(monkeypatch: MonkeyPatch):
    monkeypatch.setattr("bamboost.config.options.mpi", False)
    assert _detect_if_mpi_needed() is False


def test_detect_mpi_disabled_via_env(monkeypatch: MonkeyPatch):
    monkeypatch.setenv("BAMBOOST_MPI", "0")
    assert _detect_if_mpi_needed() is False


def test_detect_mpi_env_vars(monkeypatch: MonkeyPatch):
    monkeypatch.setattr("bamboost.config.options.mpi", True)
    monkeypatch.delenv("BAMBOOST_MPI", raising=False)
    monkeypatch.setenv("PMI_SIZE", "16")  # one of the common MPI env vars
    assert _detect_if_mpi_needed() is True


def test_detect_no_mpi(monkeypatch: MonkeyPatch):
    monkeypatch.setattr("bamboost.config.options.mpi", True)
    for key in list(os.environ):
        if key.startswith(("OMPI", "PMI", "MV2", "I_MPI", "SLURM", "MPI_")):
            monkeypatch.delenv(key, raising=False)
    assert _detect_if_mpi_needed() is False


def test_h5py_mpi_check(monkeypatch, h5py_mpi_disabled):
    with pytest.raises(RuntimeError, match="h5py was not built with MPI support"):
        _assert_h5py_has_mpi_support()


def test_h5py_mpi_check_on_import(monkeypatch: MonkeyPatch, h5py_mpi_disabled):
    # make sure MPI_ON is triggered to be True
    monkeypatch.setattr("bamboost.config.options.mpi", True)
    monkeypatch.setenv("PMI_SIZE", "16")  # ensure MPI is detected as needed

    with pytest.raises(RuntimeError):
        import bamboost.mpi as mpi

        # reload the mpi module to trigger the check
        # Should raise because h5py.get_config().mpi is False
        importlib.reload(mpi)


def test_mpi_integration_on(monkeypatch, h5py_mpi_enabled):
    monkeypatch.setattr("bamboost.config.options.mpi", True)
    monkeypatch.setenv("PMI_SIZE", "16")  # force MPI detection
    # Mock the _get_mpi_module to return a fake module and True
    # skips actual mpi4py import
    monkeypatch.setattr(
        "bamboost.mpi._get_mpi_module", lambda: ("fake-real-comm", True)
    )

    MPI, MPI_ON = get_mpi_from_env()

    assert MPI == "fake-real-comm"
    assert MPI_ON is True


def test_mpi_integration_off(monkeypatch):
    monkeypatch.setattr("bamboost.config.options.mpi", False)
    MPI, MPI_ON = get_mpi_from_env()

    assert MPI_ON is False
    assert MPI.__name__ == "bamboost.mpi.mock"  # type: ignore



from __future__ import annotations

from bamboost.core.simulation.base import Simulation as Simulation
from bamboost.core.simulation.base import SimulationWriter as SimulationWriter
from bamboost.core.simulation.types import CellType as CellType
from bamboost.core.simulation.types import FieldType as FieldType



"""Remote Access Module for Bamboost

This module provides the `Remote` class and related classes to facilitate access to remote
collections and simulations in the Bamboost framework.

Classes:
    - Remote: Represents a remote index/database, handles synchronization and workspace management.
    - RemoteCollection: Represents a collection within a remote index, supports data transfer and caching.
    - RemoteSimulation: Represents a simulation within a remote collection, supports remote data access.

Typical usage involves creating a Remote instance pointing to a remote server, listing available
collections, and synchronizing data as needed.

"""

from __future__ import annotations

import json
import subprocess
from pathlib import Path
from typing import TYPE_CHECKING, Callable, Optional, cast

from bamboost import _config, constants
from bamboost._config import config
from bamboost._logger import BAMBOOST_LOGGER
from bamboost.core.collection import Collection, _FilterKeys
from bamboost.core.simulation.base import Simulation
from bamboost.index import (
    CollectionUID,
    Index,
    create_identifier_file,
    get_identifier_filename,
)
from bamboost.index.filtering import Filter, Sorter
from bamboost.utilities import PathSet

if TYPE_CHECKING:
    from pandas import DataFrame
    from typing_extensions import Self

    from bamboost._typing import _P, StrPath
    from bamboost.mpi import Comm

log = BAMBOOST_LOGGER.getChild(__name__)


def stream_popen_output(func: Callable[_P, subprocess.Popen]) -> Callable[_P, None]:
    """Decorator to await, capture and print the output of a subprocess.Popen object."""

    def wrapper(*args, **kwargs) -> None:
        process = func(*args, **kwargs)
        assert process.stdout is not None
        for line in process.stdout:
            print(line, end="", flush=True)
        process.wait()

    return wrapper


class Remote(Index):
    """Represents a remote index/database for accessing and synchronizing collections and simulations.

    The Remote class manages connections to a remote server, handles local caching of remote data,
    and provides methods for synchronizing databases and collections using rsync. It supports
    workspace management and can list available remote databases in the local cache.

    Args:
        remote_url (str): The SSH URL of the remote server.
        comm (Optional[Comm]): Optional MPI communicator.
        workspace_path (Optional[str]): Path to the workspace on the remote server.
        workspace_name (Optional[str]): Name of the workspace.
        skip_fetch (bool): If True, skip fetching the remote database on initialization.

    Attributes:
        DATABASE_BASE_NAME (str): The base name of the database file.
        DATABASE_REMOTE_PATH (Path): Default path to the remote database.
        WORKSPACE_SPLITTER (str): String used to split remote URL and workspace name.
        id (str): Unique identifier for the remote instance.
        _remote_url (str): The SSH URL of the remote server.
        _local_path (Path): Local cache path for the remote data.
        _workspace_path (Optional[str]): Path to the workspace on the remote server.
        _workspace_name (Optional[str]): Name of the workspace.
        _remote_database_path (Path): Path to the remote database file.
        _local_database_path (Path): Path to the local cached database file.
        search_paths (PathSet): Set of search paths for collections.
        _url (str): SQLAlchemy database URL for the local cache.
        _engine: SQLAlchemy engine instance.
        _sm: SQLAlchemy sessionmaker.
        _s: SQLAlchemy session.
    """

    DATABASE_REMOTE_PATH = Path(_config._LOCAL_DIR).joinpath(
        constants.DEFAULT_DATABASE_FILE_NAME
    )
    WORKSPACE_SPLITTER = "_WS_"

    @staticmethod
    def _make_local_db_name(id: str, version: str | None = None) -> str:
        if version is not None:
            return f"{id}.{version}.sqlite"
        else:
            return f"{id}.sqlite"

    def __init__(
        self,
        remote_url: str,
        comm: Optional[Comm] = None,
        *,
        workspace_path: Optional[str] = None,
        workspace_name: Optional[str] = None,
        skip_fetch: bool = False,
    ):
        cache_dir = Path(config.paths.cacheDir)
        self._remote_url = remote_url
        self.id = (
            f"{remote_url}{self.WORKSPACE_SPLITTER}{workspace_name}"
            if workspace_name
            else remote_url
        )
        self._local_path = Path(cache_dir).joinpath(self.id)
        self._workspace_path = workspace_path
        self._workspace_name = workspace_name
        self._remote_database_path = self.DATABASE_REMOTE_PATH

        if workspace_name is not None:
            # check if the remote workspace has a meta file
            meta_file = cache_dir.joinpath(f"{self.id}.json")
            if workspace_path is not None:
                # update the meta file
                meta_file.write_text(
                    f'{{"remote_url": "{remote_url}", "workspace_path": "{workspace_path}"}}'
                )
                self._workspace_path = workspace_path
            else:
                assert meta_file.exists(), (
                    f"Workspace {workspace_name} is not found in the cache. "
                    "You must also provide the workspace path (`workspace_path=...`)."
                )
                meta = json.loads(meta_file.read_text())
                self._workspace_path = cast(str, meta["workspace_path"])

            self._remote_database_path = Path(self._workspace_path).joinpath(
                ".bamboost_cache", constants.DEFAULT_DATABASE_FILE_NAME
            )

        # Create the local path if it doesn't exist
        self._local_path.mkdir(parents=True, exist_ok=True)
        self._local_database_path = cache_dir.joinpath(
            self._make_local_db_name(self.id)
        )

        # super init
        self.search_paths = PathSet([self._local_path])
        self._url = f"sqlite:///{self._local_database_path}"

        # Fetch the remote database
        if not skip_fetch:
            # this is a blocking call to rsync
            self.fetch_remote_database()

        # Initialize the connection to the local database
        # this ensures that necessary columns are created and the database is ready for use
        self._initialize_root_process(self._url)

    def __repr__(self) -> str:
        qualname = ".".join([__name__, self.__class__.__qualname__])
        return f"<{qualname} (source={self._remote_url}, workspace={self._workspace_name})>"

    @classmethod
    def _from_id(cls, id: str) -> Remote:
        remote_url, *workspace_name = id.split(cls.WORKSPACE_SPLITTER)
        return cls(
            remote_url,
            workspace_name=workspace_name[0] if workspace_name else None,
            skip_fetch=True,
        )

    def _fetch_remote_database(
        self, db_name: str, destination: StrPath | None = None
    ) -> subprocess.Popen:
        """Fetch the remote SQL database using rsync.

        Args:
            db_name (str): The name of the database file to fetch.
            destination (StrPath): The local destination path for the database file.

        Returns:
            subprocess.Popen: The Popen object for the rsync process.
        """
        return subprocess.Popen(
            [
                "rsync",
                "-av",
                f"{self._remote_url}:{Path(_config._LOCAL_DIR).joinpath(db_name)}",
                str(destination or self._local_database_path),
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
        )

    def fetch_remote_database(
        self,
        *,
        migrate_from: str | None = None,
    ) -> None:
        """Fetch the remote SQL database."""

        if migrate_from is not None:
            from bamboost.index.versioning import Version, migrate_database

            source_version = Version.from_str(migrate_from)
            # raises ValueError if invalid

            cache_dir = Path(config.paths.cacheDir)
            tmp_dir = cache_dir.joinpath(".tmp")
            tmp_dir.mkdir(parents=True, exist_ok=True)
            tmp_db_path = tmp_dir.joinpath(
                self._make_local_db_name(self.id, migrate_from)
            )
            stream_popen_output(self._fetch_remote_database)(
                source_version.database_file_name, tmp_db_path
            )

            migrate_database(
                source_version,
                Version.latest(),
                source_db=tmp_db_path,
                destination_db=self._local_database_path,
                update=True,
            )

        else:
            stream_popen_output(self._fetch_remote_database)(
                str(constants.DEFAULT_DATABASE_FILE_NAME)
            )

    def rsync(self, source: StrPath, dest: StrPath) -> subprocess.Popen:
        """Synchronize data from the remote server to the local cache using rsync.

        Args:
            source (StrPath): The absolute source path on the remote server.
            dest (StrPath): The destination path on the local machine, relative
                to the local cache directory for this remote.

        Returns:
            subprocess.Popen: The Popen object for the rsync process.
        """
        return subprocess.Popen(
            [
                "rsync",
                "-ravh",
                f"{self._remote_url}:{source}",
                f"{self._local_path.joinpath(dest)}",
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
        )

    @classmethod
    def list(cls) -> list[Remote]:
        """List all remote databases in the cache."""
        # find all sqlite files in the cache directory
        return [
            Remote._from_id(name.stem)
            for name in _config.CACHE_DIR.iterdir()
            if name.is_file() and name.suffix == ".sqlite"
        ]

    def __getitem__(self, key: str) -> RemoteCollection:
        uid = key.split(" - ")[0]
        return RemoteCollection(uid, remote=self)

    def _ipython_key_completions_(self) -> list[str]:
        return [coll.uid for coll in self._get_collections()]

    def get_local_path(self, collection_uid: str) -> Path:
        return self._local_path.joinpath(collection_uid)


class RemoteCollection(Collection):
    _index: Remote

    def __init__(
        self,
        uid: str,
        remote: Remote,
        *,
        filter: Filter | None = None,
        sorter: Sorter | None = None,
    ):
        self.uid = CollectionUID(uid)
        self._index = remote  # pyright: ignore[reportIncompatibleVariableOverride]

        # A key store with completion of all the parameters and metadata keys
        self.k = _FilterKeys(self)

        # Resolve the path (this updates the index if necessary)
        self.path = remote._local_path.joinpath(uid)
        self.remote_path = cast(Path, remote._get_collection_path(self.uid))

        # Create the diretory for the collection if necessary
        self.path.mkdir(parents=True, exist_ok=True)

        # Check if identifier file exists (and create it if necessary)
        if not self.path.joinpath(get_identifier_filename(uid=self.uid)).exists():
            create_identifier_file(self.path, self.uid)

        self._filter = filter
        self._sorter = sorter

    def _repr_html_(self) -> str:
        """HTML repr for ipython/notebooks, using jinja2 for templating."""
        import pkgutil

        from jinja2 import Template

        html_string = pkgutil.get_data("bamboost", "_repr/manager.html").decode()  # type: ignore
        icon = pkgutil.get_data("bamboost", "_repr/icon.txt").decode()  # type: ignore

        template = Template(html_string)

        return template.render(
            icon=icon,
            db_path=f"<a href={self.path.as_posix()}>{self.path}</a>",
            db_uid=f"{self.uid} ({self._index._remote_url})",
            db_size=len(self),
            _filter=self._filter,
            _sort=self._sorter,
        )

    def __getitem__(self, name_or_idx: str | int) -> RemoteSimulation:
        if isinstance(name_or_idx, int):
            name: str = self.df.iloc[name_or_idx]["name"]
        else:
            name = name_or_idx
        return RemoteSimulation(
            name, self.path, collection_uid=self.uid, index=self._index, comm=self._comm
        )

    @property
    def df(self) -> DataFrame:
        df = super().df

        # Add a cached column at the second position
        cached_col = df["name"].isin(self.get_cached_simulation_names())
        df.insert(1, "cached", cached_col)  # pyright: ignore[reportArgumentType]
        return df

    def get_cached_simulation_names(self) -> list[str]:
        return [str(i.name) for i in self.path.iterdir() if i.is_dir()]

    def rsync(self, name: Optional[str] = None) -> Self:
        """Transfer data using rsync. Wait for the process to finish and return
        self.

        Args:
            name: The simulation name of the simulation to be transferred. If None, all
                simulations are synced.
        """
        if name:
            source = self.remote_path.joinpath(name)
            dest = self.path.joinpath(name)
        else:
            # source will be past to rsync as <remote_url>:<source>
            # source should be the glob: <remote_path>/*
            source = self.remote_path.joinpath("*")
            dest = self.path

        stream_popen_output(self._index.rsync)(source, dest)
        return self


class RemoteSimulation(Simulation):
    _index: Remote

    def __init__(
        self,
        name: str,
        parent: StrPath,
        collection_uid: CollectionUID,
        index: Remote,
        comm: Optional[Comm] = None,
        **kwargs,
    ):
        # if there is no local path, sync it from remote
        # assumes that this is the first time accessing the simulation
        # initialization of base Simulation requires the path to exist
        # (including the data.h5 file)
        path = index.get_local_path(collection_uid).joinpath(name)
        self.remote_path: Path = index._get_collection_path(collection_uid).joinpath(
            name
        )
        if not path.exists() or not path.joinpath("data.h5").exists():
            log.info(
                f"Local simulation '{name}' not found in collection '{collection_uid}'. "
                "Syncing from remote..."
            )
            stream_popen_output(index.rsync)(self.remote_path, path.parent)

        # call super init which requires the path to exist otherwise it will raise FileNotFoundError
        super().__init__(
            name,
            parent,
            index=index,
            comm=comm,
            collection_uid=collection_uid,
            **kwargs,
        )

    @property
    def uid(self) -> str:
        return f"ssh://{self._index._remote_url}/{super().uid}"

    @property
    def parameters(self) -> dict:
        return self._orm.parameter_dict

    @property
    def metadata(self) -> dict:
        return self._orm.as_dict(include_parameters=False, include_links=False)

    def rsync(self) -> Self:
        """Sync the simulation data with the remote server."""
        stream_popen_output(self._index.rsync)(self.remote_path, self.path.parent)
        return self



from typing import Optional

import typer

from bamboost.cli.common import console

app_config = typer.Typer(
    no_args_is_help=True,
    pretty_exceptions_show_locals=False,
    help="Print or edit the configuration.",
)


@app_config.command()
def print(
    dir: Optional[str] = typer.Option(
        None, "--dir", "-d", help="Directory to show the config for."
    ),
):
    """Show the active configuration."""
    if dir:
        from bamboost._config import _Config

        config = _Config(dir)
    else:
        from bamboost import config

    console.print(config)


@app_config.command()
def show():
    """Open the global config file."""
    import os

    from bamboost._config import CONFIG_FILE

    os.system(f"${{EDITOR:-vim}} {CONFIG_FILE}")



"""Module for handling time series data in HDF5 files. The notion of time is used to
describe Series, but it can be used for any data that changes with a single parameter.

The default time series is stored at `/data` in the HDF5 file. You can create additional
series' with `bamboost.core.simulation.base.Simulation.create_series`.

This module provides classes for managing time series data stored in HDF5 files:
- Series: Main class for managing a series, including fields and global values
- FieldData: Handles a particular field (e.g., nodal or element data) and its timesteps
- GlobalData: Manages the global data that varies with time. This refers to data that is
  not tied to a mesh.
- StepWriter: Helper class for writing data at specific timesteps

The data is organized hierarchically in the HDF5 file with separate sections for
field data and global/scalar data. Fields typically represent spatial data like
displacements or stresses, while globals are used for scalar quantities like
energy or convergence metrics.
"""

from __future__ import annotations

import pkgutil
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, Iterable, Optional, Union, overload

import h5py
import numpy as np

import bamboost
from bamboost import constants
from bamboost._logger import BAMBOOST_LOGGER
from bamboost._typing import _MT, ArrayLike, Mutable
from bamboost.constants import (
    DEFAULT_MESH_NAME,
    PATH_DATA,
    RELATIVE_PATH_FIELD_DATA,
    RELATIVE_PATH_SCALAR_DATA,
)
from bamboost.core.hdf5.file import (
    FileMode,
    H5Object,
    WriteInstruction,
    mutable_only,
    with_file_open,
)
from bamboost.core.hdf5.hdf5path import HDF5Path
from bamboost.core.hdf5.ref import Dataset, Group, H5Reference, InvalidReferenceError
from bamboost.core.simulation.types import FieldType

if TYPE_CHECKING:
    import pandas as pd

    from bamboost.core.simulation.base import _Simulation

log = BAMBOOST_LOGGER.getChild(__name__)


class NotASeriesError(ValueError):
    def __init__(self, path: str):
        super().__init__(f"Path {path} exists, but is not a Series.")


class Series(H5Reference[_MT]):
    _obj: h5py.Group

    def __init__(self, simulation: _Simulation[_MT], path: str = PATH_DATA):
        super().__init__(path, simulation._file)

        # if this is not tagged a series, we raise an error
        if not self.attrs.get(".series"):
            raise NotASeriesError(path)
        self._field_instances: dict[str, FieldData[_MT]] = {}

    def __len__(self) -> int:
        try:
            return self._values().shape[0]
        except (InvalidReferenceError, KeyError):
            return 0

    @overload
    def __getitem__(self, key: tuple[()]) -> list[FieldData[_MT]]: ...
    @overload
    def __getitem__(
        self, key: Union[list[str], tuple[str], set[str]]
    ) -> list[FieldData[_MT]]: ...
    @overload
    def __getitem__(self, key: str) -> FieldData[_MT]: ...
    def __getitem__(self, key):
        if isinstance(key, str):
            return self.get_field(key)
        if isinstance(key, tuple) and len(key) == 0:
            return self.get_fields()
        # else the key is a iterable of strings
        return [self.get_field(k) for k in key]

    def _ipython_key_completions_(self) -> list[HDF5Path]:
        return self.get_field_names()

    @with_file_open(FileMode.READ)
    def _repr_html_(self):
        """Repr showing the content of the group."""
        from jinja2 import Template

        html_template = pkgutil.get_data(
            bamboost.__name__, "_repr/series.html"
        ).decode()  # type: ignore
        icon = pkgutil.get_data(bamboost.__name__, "_repr/icon.txt").decode()  # type: ignore
        template = Template(html_template)

        return template.render(
            uid=Path(self._file._filename).parent.name,
            name=self._path,
            icon=icon,
            version=bamboost.__version__,
            attrs=self.attrs,
            globals=[(k, self.globals[k].shape) for k in self.globals.keys()],
            fields=self.get_field_names(),
            size=len(self),
        )

    @property
    def last_step(self) -> Union[int, None]:
        if not hasattr(self, "_last_step"):
            try:
                self._last_step = self._values().shape[0] - 1
            except KeyError:
                self._last_step = None
        return self._last_step

    @last_step.setter
    def last_step(self, value: int):
        self._last_step = value

    def get_field_names(self) -> list[HDF5Path]:
        """Return all the name of the fields in the series."""
        if not hasattr(self, "__fields_group"):
            self.__fields_group = Group(
                self._path.joinpath(RELATIVE_PATH_FIELD_DATA), self._file
            )
        return list(self.__fields_group._group_map.children())

    def get_field(self, name: str) -> FieldData[_MT]:
        """Get a field by name.

        Args:
            name: The name of the field.
        """
        if name not in self._field_instances:
            self._field_instances[name] = FieldData(self, name)
        return self._field_instances[name]

    def get_fields(self, *glob: str) -> list[FieldData[_MT]]:
        """Get multiple fields by name or glob pattern. If no arguments are given, all
        fields are returned.

        Args:
            glob: A list of glob patterns to filter the field names.
        """
        if not glob:
            return [self.get_field(name) for name in self.get_field_names()]

        import fnmatch

        matching_fields = set()
        for g in glob:
            matching_fields.update(fnmatch.filter(self.get_field_names(), g))
        return [self.get_field(name) for name in matching_fields]

    @cached_property
    def globals(self) -> GlobalData[_MT]:
        return GlobalData(self)

    @property
    @with_file_open(FileMode.READ)
    def values(self) -> np.ndarray:
        """Return the values of the series. In the default time series, this returns the
        time values of the steps."""
        try:
            return self._values()[:]
        except InvalidReferenceError:
            return np.array([])

    def _values(self) -> Dataset[_MT]:
        return super().__getitem__((constants.DS_NAME_TIMESTEPS, Dataset))

    @mutable_only
    def require_step(
        self: Series[Mutable],
        value: float = np.nan,
        step: Optional[int] = None,
    ) -> StepWriter:
        """Create a new step in the series. If the step is not given, we append one.

        Args:
            value: The value of the step. This is typically the time value.
            step: The step number. If not given, a step is appended after the last one.
        """
        if step is None:
            step = self.last_step + 1 if self.last_step is not None else 0

        # store the timestep if given
        with self.suspend_immediate_write():
            self._store_value(step, value)
            self.last_step = step

        return StepWriter(self, step)

    @mutable_only
    def _store_value(self: Series[Mutable], step: int, time: float) -> None:
        def _write_instruction():
            # require the dataset for the timesteps
            dataset = self._obj.require_dataset(
                constants.DS_NAME_TIMESTEPS,
                shape=(step + 1,),
                dtype=np.float64,
                chunks=True,
                maxshape=(None,),
                fillvalue=np.nan,
            )

            # resize the dataset and store the time
            new_size = max(step + 1, dataset.shape[0])
            log.debug(f"Resizing dataset {dataset.name} to {new_size}")
            dataset.resize(new_size, axis=0)

            log.debug(f"Storing timestep {time} for step {step}")
            dataset[step] = time

        self.post_write_instruction(_write_instruction)


class StepWriter(H5Object[Mutable]):
    """A class to write data for a specific step in a series.

    Args:
        series: The series to which the step belongs.
        step: The step number.
    """

    def __init__(self, series: Series[Mutable], step: int):
        super().__init__(series._file)
        self._series = series
        """The series to which the step belongs."""
        self._step = step
        """The step number."""

    def add_field(
        self,
        name: str,
        data: ArrayLike,
        *,
        mesh_name: str = DEFAULT_MESH_NAME,
        field_type: FieldType = FieldType.NODE,
    ) -> None:
        """Add a field to the step.

        Args:
            name: The name of the field.
            data: The data for the field.
            mesh_name: The name of the mesh to which the field belongs.
            field_type: The type of the field (default: FieldType.NODE). This is only
                relevant for XDMF writing.
        """
        field = self._series.get_field(name)
        field.require_self()
        field.add_numerical_dataset(
            str(self._step),
            data,
            file_map=True,
            attrs={"mesh": mesh_name, "type": field_type.value},
        )
        log.debug(f"Added field {name} for step {self._step}")

    def add_fields(
        self,
        fields: dict[str, ArrayLike],
        mesh_name: str = DEFAULT_MESH_NAME,
        field_type: FieldType = FieldType.NODE,
    ) -> None:
        """Add multiple fields to the step.

        Args:
            fields: A dictionary of field names and their data.
            mesh_name: The name of the mesh to which the fields belong.
            field_type: The type of the fields (default: FieldType.NODE). This is only
                relevant for XDMF writing.
        """
        with self._file.open(FileMode.APPEND, driver="mpio"):
            for name, data in fields.items():
                self.add_field(name, data, mesh_name=mesh_name, field_type=field_type)

    def add_scalar(self, name: str, data: Union[int, float, Iterable]) -> None:
        """Add a scalar to the step. Scalar data is typically a single value or a small
        array. The shape must be consistent across all steps.

        Args:
            name: The name of the scalar.
            data: The data for the scalar.

        Raises:
            ValueError: If the shape of the data is not consistent with the existing data.
        """

        @dataclass
        class AddScalarInstruction(WriteInstruction):
            series: Series
            name: str
            data_arr: np.ndarray
            step: int

            def __call__(self):
                log.info(f"Adding scalar {self.name} for step {self.step}")
                dataset = self.series.globals.require_dataset(
                    self.name,
                    shape=(1, *self.data_arr.shape),
                    dtype=float,
                    maxshape=(None, *self.data_arr.shape),
                    chunks=True,
                    fillvalue=np.nan,
                )
                new_size = max(self.step + 1, dataset.shape[0])
                if new_size > dataset.shape[0]:
                    log.info(f"Resizing dataset {dataset.name} to {new_size}")
                dataset.resize(new_size, axis=0)
                dataset[self.step] = self.data_arr

        self.post_write_instruction(
            AddScalarInstruction(self._series, name, np.array(data), self._step)
        )

    def add_scalars(self, scalars: dict[str, Union[int, float, Iterable]]) -> None:
        """Add multiple scalars to the step. See `add_scalar` for more information.

        Args:
            scalars: A dictionary of scalar names and their data.
        """
        with self.suspend_immediate_write():
            for name, data in scalars.items():
                self.add_scalar(name, data)

        self._file.single_process_queue.apply()


class FieldData(Group[_MT]):
    _parent: Series[_MT]
    name: str

    def __init__(self, series: Series[_MT], name: str):
        super().__init__(
            series._path.joinpath(RELATIVE_PATH_FIELD_DATA, name), series._file
        )
        self._parent = series
        self.name = name

    def __getitem__(
        self, key: Union[int, slice, tuple[slice | int, ...]]
    ) -> np.ndarray:
        """Get data of the field for a specific step or steps using standard slice
        notation. First index is the step number.
        """
        if isinstance(key, Iterable):
            step = key[0]
            rest = key[1:]
        else:
            step = key
            rest = ()

        with self._file.open(FileMode.READ):
            if isinstance(step, int):
                step_positive = self._handle_negative_index(step)
                try:
                    return self._obj[str(step_positive)][rest]  # type: ignore
                except KeyError:
                    raise IndexError(
                        f"Index ({step_positive}) out of range for (0-{self._parent.last_step})"
                    )
            else:
                try:
                    return np.array(
                        [self._obj[i][rest] for i in self._slice_step(step)]  # pyright: ignore[reportIndexIssue]
                    )
                except ValueError:  # The data has an inhomogeneous shape (different shape at different steps)
                    return np.array(
                        [self._obj[i][rest] for i in self._slice_step(step)],  # pyright: ignore[reportIndexIssue]
                        dtype=object,
                    )

    def at(self, step: int) -> Dataset[_MT]:
        """Get the dataset for a specific step without reading the data itself.

        Args:
            step: The step number.
        """
        step_positive = self._handle_negative_index(step)
        try:
            return self.new(
                self._path.joinpath(str(step_positive)), self._file, Dataset
            )
        except KeyError:
            raise IndexError(
                f"Index ({step_positive}) out of range for (0-{self._parent.last_step})"
            )

    def _handle_negative_index(self, index: int) -> int:
        if index < 0:
            return (self._parent.last_step or 0) + index + 1
        return index

    def _slice_step(self, step: slice) -> list[str]:
        indices = [
            str(i) for i in range(*step.indices((self._parent.last_step or 0) + 1))
        ]
        return indices


class GlobalData(Group[_MT]):
    def __init__(self, series: Series[_MT]):
        super().__init__(series._path.joinpath(RELATIVE_PATH_SCALAR_DATA), series._file)
        self._series = series

    def __getitem__(self, key: str) -> Dataset[_MT]:
        return super().__getitem__((key, Dataset[_MT]))

    @property
    def df(self) -> pd.DataFrame:
        from pandas import DataFrame

        return DataFrame(
            {"values": self._series.values} | {k: list(v[:]) for k, v in self.items()}
        )

    def get(self, *glob: str) -> tuple[tuple[np.ndarray, ...], tuple[str, ...]]:
        """Get the data for many scalars by name or glob pattern. If no arguments are
        given, all scalars are read and returned.

        Args:
            glob: A list of glob patterns to filter the scalar names.
        """
        if not glob:
            dataset_names = self.datasets()
            return tuple(self.__getitem__(name).array for name in dataset_names), tuple(
                map(str, dataset_names)
            )

        import fnmatch

        matching_datasets = set()
        for g in glob:
            matching_datasets.update(fnmatch.filter(self.datasets(), g))
        matching_datasets = tuple(matching_datasets)
        return tuple(
            self.__getitem__(name).array for name in matching_datasets
        ), matching_datasets



import os
import shutil
import tempfile
from pathlib import Path

import pytest

from bamboost import Collection, Index, config
from bamboost.mpi import Communicator


def pytest_sessionstart(session):
    """Setup tmp config directory."""
    # Use tempdir for testing
    tempdir = tempfile.mkdtemp()
    tempdir = Path(tempdir)
    config.paths.localDir = tempdir
    config.paths.cacheDir = tempdir.joinpath("cache")
    # Disable MPI for testing
    config.options.mpi = False
    # Use in-memory database for testing
    config.index.databaseFile = ":memory:"

    # Create config files if they don't exist
    os.makedirs(config.paths.localDir, exist_ok=True)


def pytest_sessionfinish(session, exitstatus):
    """Remove tmp config directory again."""
    shutil.rmtree(config.paths.localDir)


@pytest.fixture(scope="module")
def tmp_path_module(tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp("data")
    tmp_path = Communicator._active_comm.bcast(tmp_path, root=0)
    yield tmp_path


@pytest.fixture(scope="module")
def tmp_collection(tmp_path_module: Path):
    yield Collection(
        path=tmp_path_module.joinpath("tmp_collection"),
        index_instance=Index.default,
    )


@pytest.fixture
def tmp_collection_burn(tmp_path: Path):
    tmp_path = Communicator._active_comm.bcast(tmp_path, root=0)
    yield Collection(
        path=tmp_path,
        index_instance=Index.default,
    )


@pytest.fixture(scope="module")
def test_collection(tmp_collection: Collection):
    tmp_collection.add(
        "testsim1",
        parameters={
            "first_name": "John",
            "age": 20,
            "list": [1, 2, 3],
            "dict": {"a": 1, "b": 2},
        },
    )
    tmp_collection.add(
        "testsim2",
        parameters={
            "first_name": "Jane",
            "age": 30,
            "list": [4, 5, 6],
            "dict": {"c": 3, "d": 4},
        },
    )
    tmp_collection.add(
        "testsim3",
        parameters={
            "first_name": "Jack",
            "age": 40,
            "list": [7, 8, 9],
            "dict": {"e": 5, "f": 6},
        },
    )
    yield tmp_collection



import sqlite3
from contextlib import contextmanager


class FastIndexQuery:
    def __init__(self):
        from bamboost._config import config

        self.db_path = config.index.databaseFile

    @contextmanager
    def connection(self):
        try:
            self._conn = sqlite3.connect(self.db_path)
            self._cursor = self._conn.cursor()
            yield self._cursor
        finally:
            self._conn.close()

    def query(self, query: str, *args) -> list[tuple]:
        with self.connection() as cursor:
            cursor.execute(query, *args)
            return cursor.fetchall()


INDEX = FastIndexQuery()



from __future__ import annotations

from collections.abc import ItemsView, KeysView, Mapping, MutableMapping
from typing import TYPE_CHECKING, Callable, Generator, Iterator, Type, Union

import h5py

from bamboost.core.hdf5.hdf5path import HDF5Path

if TYPE_CHECKING:
    from bamboost.core.hdf5.file import HDF5File

_VT_filemap = Type[Union[h5py.Group, h5py.Dataset]]


class KeysViewHDF5(KeysView):
    def __str__(self) -> str:
        return "<KeysViewHDF5 {}>".format(list(self))

    __repr__ = __str__


class _FileMapMixin(Mapping[HDF5Path, _VT_filemap]):
    def keys(self) -> KeysViewHDF5:
        return KeysViewHDF5(self)

    def datasets(self) -> tuple[str, ...]:
        return tuple(k for k, v in self.items() if v is h5py.Dataset)

    def groups(self) -> tuple[str, ...]:
        return tuple(k for k, v in self.items() if v is h5py.Group)

    def _ipython_key_completions_(self):
        return self.keys()

    def _repr_pretty_(self, p, _cycle):
        p.pretty(dict(self))

    def __repr__(self):
        return f"{type(self).__name__}({dict(self)})"


class FileMap(MutableMapping[HDF5Path, _VT_filemap], _FileMapMixin):
    _valid: bool

    def __init__(self, file: HDF5File):
        self._file = file
        self._dict: dict[HDF5Path, _VT_filemap] = {}
        self.valid = False

    def __getitem__(self, key: str, /) -> _VT_filemap:
        return self._dict[HDF5Path(key)]

    def __setitem__(self, key: str, value: _VT_filemap) -> None:
        self._dict[HDF5Path(key)] = value

    def __delitem__(self, key: str) -> None:
        self._dict.pop(HDF5Path(key))

    def __iter__(self) -> Iterator[HDF5Path]:
        return iter(self._dict)

    def __len__(self) -> int:
        return len(self._dict)

    def populate(self, *, exclude_numeric: bool = False) -> None:
        """Assumes the file is open."""

        # visit all groups and datasets to cache them
        def cache_items(name, _obj):
            path = HDF5Path(name)
            if exclude_numeric and path.basename.isdigit():
                return
            self._dict[path] = type(_obj)

        self._file.visititems(cache_items)
        self.valid = True

    def invalidate(self) -> None:
        self.valid = False


class FilteredFileMap(MutableMapping[HDF5Path, _VT_filemap], _FileMapMixin):
    def __init__(self, file_map: FileMap, parent: str) -> None:
        self.parent = HDF5Path(parent)
        self.file_map = file_map

    @property
    def valid(self) -> bool:
        return self.file_map.valid

    @valid.setter
    def valid(self, value: bool) -> None:
        self.file_map.valid = value

    def __getitem__(self, key: str, /):
        return self.file_map[self.parent.joinpath(key)]

    def __setitem__(self, key: str, value):
        self.file_map[self.parent.joinpath(key)] = value

    def __delitem__(self, key):
        del self.file_map[self.parent.joinpath(key)]

    def children(self) -> Iterator[HDF5Path]:
        return filter(lambda path: "/" not in path, self)

    def children_groups(self) -> Iterator[HDF5Path]:
        return filter(lambda path: self[path] is h5py.Group, self.children())

    def children_datasets(self) -> Iterator[HDF5Path]:
        return filter(lambda path: self[path] is h5py.Dataset, self.children())

    def __iter__(self):
        return map(
            lambda x: HDF5Path(x).relative_to(self.parent),
            filter(
                lambda x: HDF5Path(x).is_child_of(self.parent),
                self.file_map,
            ),
        )

    def __len__(self):
        return sum(1 for _ in self.__iter__())



# Simulation specific constants
UID_SEPARATOR = ":"
HDF_DATA_FILE_NAME = "data.h5"
XDMF_FILE_NAME = "data.xdmf"
RUN_FILE_NAME = "run.sh"

PATH_DATA = "data"
RELATIVE_PATH_FIELD_DATA = "fields"
RELATIVE_PATH_SCALAR_DATA = "globals"
PATH_PARAMETERS = ".parameters"
PATH_MESH = ".mesh"
PATH_LINKS = ".links"
PATH_USERDATA = ".userdata"

DS_NAME_TIMESTEPS = ".map"

DEFAULT_MESH_NAME = "default"

# SQLModel constants
TABLENAME_COLLECTIONS = "collections"
TABLENAME_SIMULATIONS = "simulations"
TABLENAME_PARAMETERS = "parameters"
TABLENAME_SIMULATION_LINKS = "simulation_links"

# Config constants
DEFAULT_CONFIG_FILE_NAME = "config.toml"
DEFAULT_DATABASE_FILE_NAME = "bamboost.0.11.sqlite"



"""SQLAlchemy Core schema definitions and helper utilities for the BAMBOOST index database."""

from __future__ import annotations

import dataclasses
import json
from dataclasses import dataclass, field, fields
from datetime import datetime
from typing import TYPE_CHECKING, Any, Callable, Iterable, Mapping, Sequence

from sqlalchemy import (
    JSON,
    Boolean,
    Column,
    DateTime,
    Engine,
    ForeignKey,
    Integer,
    MetaData,
    RowMapping,
    String,
    Table,
    UniqueConstraint,
    delete,
    select,
    tuple_,
)
from sqlalchemy.dialects.sqlite import insert
from sqlalchemy.orm import Session
from sqlalchemy.sql.dml import Delete, Insert, ReturningInsert
from typing_extensions import Self

from bamboost import constants
from bamboost._logger import BAMBOOST_LOGGER
from bamboost.constants import (
    TABLENAME_COLLECTIONS,
    TABLENAME_PARAMETERS,
    TABLENAME_SIMULATION_LINKS,
    TABLENAME_SIMULATIONS,
)
from bamboost.core.utilities import flatten_dict
from bamboost.index.filtering import Filter, Sorter
from bamboost.index.uids import SimulationUID

if TYPE_CHECKING:
    from pandas import DataFrame

log = BAMBOOST_LOGGER.getChild(__name__)

# ------------------------------------------------
# Dataclasses representing the loaded records
# ------------------------------------------------


@dataclass(frozen=True)
class ParameterRecord:
    id: int
    simulation_id: int
    key: str
    value: Any

    @classmethod
    def from_mapping(cls, row: RowMapping) -> ParameterRecord:
        return cls(
            id=row["id"],
            simulation_id=row["simulation_id"],
            key=row["key"],
            value=row["value"],
        )


@dataclass(frozen=True)
class LinkRecord:
    id: int
    source_id: int
    target_id: int
    source_uid: SimulationUID
    name: str

    @classmethod
    def from_mapping(cls, row: RowMapping) -> LinkRecord:
        return cls(
            id=row["id"],
            source_id=row["source_id"],
            target_id=row["target_id"],
            source_uid=SimulationUID(row["source_coll_uid"], row["source_name"]),
            name=row["name"],
        )


@dataclass(frozen=True)
class SimulationRecord:
    id: int
    collection_uid: str
    name: str
    created_at: datetime
    modified_at: datetime
    description: str | None
    status: str
    submitted: bool
    tags: list[str] = field(default_factory=list)
    parameters: list[ParameterRecord] = field(default_factory=list)

    links: dict[str, SimulationUID] = field(default_factory=dict)

    @classmethod
    def from_mapping(
        cls,
        row: RowMapping,
        parameters: list[ParameterRecord],
        links: dict[str, SimulationUID],
    ) -> SimulationRecord:
        return cls(
            id=row["id"],
            collection_uid=row["collection_uid"],
            name=row["name"],
            created_at=row["created_at"],
            modified_at=row["modified_at"],
            description=row.get("description"),
            tags=list(row.get("tags") or []),
            status=row.get("status", ""),
            submitted=bool(row.get("submitted", False)),
            parameters=list(parameters or []),
            links=links or {},
        )

    @property
    def parameter_dict(self) -> dict[str, Any]:
        return {parameter.key: parameter.value for parameter in self.parameters}

    def as_dict(
        self,
        standalone: bool = True,
        include_parameters: bool = True,
        include_links: bool = True,
    ) -> dict[str, Any]:
        if standalone:
            output_fields = ("id", "collection_uid", "name", "created_at", "modified_at", "description", "tags", "status", "submitted")  # fmt: off
        else:
            # When not standalone (e.g. for DataFrames), we exclude the internal 'id'
            output_fields = ("name", "created_at", "description", "tags", "status", "submitted")  # fmt: off
        if include_links:
            output_fields += ("links",)

        data = {field: getattr(self, field) for field in output_fields}
        if include_links and self.links:
            # Convert SimulationUID objects to strings for the dictionary
            data["links"] = {k: str(v) for k, v in self.links.items()}

        if include_parameters:
            data.update(self.parameter_dict)
        return data


@dataclass(frozen=False)
class CollectionMetadata:
    """Collection metadata container. Can load/save from/to dicts, storing unknown
    fields in an extras dictionary.
    """

    uid: str
    """Unique identifier of the collection."""
    created_at: datetime | None = field(default=None)
    """Creation timestamp."""
    tags: list[str] = field(default_factory=list)
    """List of tags associated with the collection."""
    aliases: list[str] = field(default_factory=list)
    """List of aliases for the collection."""
    author: str | dict | None = field(default=None)
    """Author information, can be a string or a dictionary."""
    description: str | None = field(default=None)
    """Completely optional description of the collection."""

    _extras: dict[str, Any] = field(default_factory=dict, init=False, repr=False)

    def __post_init__(self):
        self._fields = {f.name for f in fields(self) if f.init}  # ignore _extras

    def to_dict(self) -> dict[str, Any]:
        d = {k: getattr(self, k) for k in self._fields}
        d = {k: v for k, v in d.items() if v is not None}  # remove None values
        d.update(self._extras)
        return d

    @classmethod
    def from_dict(cls, data: dict[str, Any], *_) -> Self:
        """Create an instance from a dictionary, storing unknown fields in the extras
        dictionary.

        Args:
            data: The input dictionary.
        """
        _fields = {f.name for f in fields(cls) if f.init}  # ignore _extras
        known = {k: v for k, v in data.items() if k in _fields}
        extras = {k: v for k, v in data.items() if k not in _fields}

        obj = cls(**known)
        obj._extras = extras
        return obj


@dataclass(frozen=False)
class CollectionRecord(CollectionMetadata):
    """Collection record, including metadata and associated simulations. Used for exposing
    (read only) a record in the collection table in the sql database.
    """

    path: str = field(
        default=""
    )  # default to empty string for compatibility, it should always be set
    """Path to the collection on disk. Guaranteed to be not null."""
    simulations: list[SimulationRecord] | property = field(
        default_factory=list, repr=False
    )
    """List of simulations associated with the collection."""

    @classmethod
    def from_mapping(
        cls, row: RowMapping, simulations: list[SimulationRecord]
    ) -> CollectionRecord:
        return CollectionRecord(
            uid=row["uid"],
            path=row["path"],
            created_at=row.get("created_at"),
            description=row.get("description", ""),
            tags=list(row.get("tags") or []),
            aliases=list(row.get("aliases") or []),
            author=row.get("author"),
            simulations=simulations,
        )

    @property
    def parameters(self) -> list[ParameterRecord]:
        return [
            parameter
            for simulation in self.simulations
            for parameter in simulation.parameters
        ]

    @property
    def links(self) -> dict[str, dict[str, SimulationUID]]:
        """Dictionary of links for each simulation in the collection."""
        return {sim.name: sim.links for sim in self.simulations if sim.links}

    def get_parameter_keys(self) -> tuple[list[str], list[int]]:
        unique = list({parameter.key for parameter in self.parameters})
        counts = [
            sum(1 for parameter in self.parameters if parameter.key == key)
            for key in unique
        ]
        return unique, counts

    def to_pandas(self, flatten: bool = True) -> "DataFrame":
        import pandas as pd

        records = [
            simulation.as_dict(standalone=False, include_links=False)
            for simulation in self.simulations
        ]
        if flatten:
            records = [flatten_dict(record) for record in records]
        return pd.DataFrame.from_records(records)

    def filtered(self, filter: Filter | None, sorter: Sorter | None) -> Self:
        # TODO: this method is not efficient!
        if (filter is None) and (sorter is None):
            return self

        df = self.to_pandas()
        if filter is not None:
            df = filter.apply(df)
        if sorter is not None:
            tmp_map = {
                sim.name: sim
                for sim in self.simulations
                if sim.name in df["name"].values
            }
            df = sorter.apply(df)  # type: ignore
            simulations = [tmp_map[name] for name in df["name"].values]
        else:
            simulations = [
                sim for sim in self.simulations if sim.name in df["name"].values
            ]

        return dataclasses.replace(self, simulations=simulations)


# ------------------------------------------------
# SQL table definitions
# ------------------------------------------------
metadata = MetaData()

collections_table = Table(
    TABLENAME_COLLECTIONS,
    metadata,
    Column("uid", String, primary_key=True),
    Column("path", String, nullable=False),
    Column("created_at", DateTime, nullable=True),
    Column("description", String, nullable=False, default="", server_default=""),
    Column("tags", JSON, nullable=False, default=list, server_default="[]"),
    Column("aliases", JSON, nullable=False, default=list, server_default="[]"),
    Column("author", JSON, nullable=True, default=None, server_default="null"),
)

simulations_table = Table(
    TABLENAME_SIMULATIONS,
    metadata,
    Column("id", Integer, primary_key=True, autoincrement=True),
    Column(
        "collection_uid",
        String,
        ForeignKey(f"{TABLENAME_COLLECTIONS}.uid", ondelete="CASCADE"),
    ),
    Column("name", String, nullable=False),
    Column("created_at", DateTime, nullable=False, default=datetime.now),
    Column("modified_at", DateTime, nullable=False, default=datetime.now),
    Column("description", String, nullable=True),
    Column("tags", JSON, nullable=False, default=list, server_default="[]"),
    Column(
        "status",
        String,
        nullable=False,
        default="initialized",
        server_default="initialized",
    ),
    Column("submitted", Boolean, nullable=False, default=False, server_default="0"),
    UniqueConstraint("collection_uid", "name", name="uix_collection_name"),
)

simulation_links_table = Table(
    TABLENAME_SIMULATION_LINKS,
    metadata,
    Column("id", Integer, primary_key=True, autoincrement=True),
    Column(
        "source_id",
        Integer,
        ForeignKey(f"{TABLENAME_SIMULATIONS}.id", ondelete="CASCADE"),
        nullable=False,
    ),
    Column(
        "target_id",
        Integer,
        ForeignKey(f"{TABLENAME_SIMULATIONS}.id", ondelete="CASCADE"),
        nullable=False,
    ),
    Column("name", String, nullable=False),
    UniqueConstraint("source_id", "name", name="uix_source_link_name"),
)

parameters_table = Table(
    TABLENAME_PARAMETERS,
    metadata,
    Column("id", Integer, primary_key=True, autoincrement=True),
    Column(
        "simulation_id",
        Integer,
        ForeignKey(f"{TABLENAME_SIMULATIONS}.id", ondelete="CASCADE"),
        nullable=False,
    ),
    Column("key", String, nullable=False),
    Column("value", JSON, nullable=False),
    UniqueConstraint("simulation_id", "key", name="uix_simulation_key"),
)


def create_all(engine: Engine) -> None:
    """Create all tables defined in this module."""

    metadata.create_all(engine, checkfirst=True)
    with engine.begin() as connection:
        for table in (
            collections_table,
            simulations_table,
            parameters_table,
            simulation_links_table,
        ):
            existing_columns = {
                row[1]
                for row in connection.exec_driver_sql(
                    f"PRAGMA table_info({table.name})"
                )
            }

            for column in table.columns:
                if column.name in existing_columns:
                    continue
                connection.exec_driver_sql(
                    f"ALTER TABLE {table.name} ADD COLUMN "
                    f"{_column_sql_for_alter(engine, column)}"
                )


def _column_sql_for_alter(engine: Engine, column: Column[Any]) -> str:
    """Build a SQLite-safe column definition for ALTER TABLE ADD COLUMN."""
    type_sql = column.type.compile(dialect=engine.dialect)
    clauses = [column.name, str(type_sql)]

    default_sql = _server_default_sql(column)
    # SQLite refuses ALTER TABLE ADD COLUMN for NOT NULL columns without a default.
    # Keep migration permissive for legacy schemas by only enforcing NOT NULL when
    # a server default is available.
    if not column.nullable and default_sql is not None:
        clauses.append("NOT NULL")

    if default_sql is not None:
        clauses.append(f"DEFAULT {default_sql}")

    return " ".join(clauses)


def _server_default_sql(column: Column[Any]) -> str | None:
    if column.server_default is None:
        return None

    arg = column.server_default.arg  # type: ignore[unresolved-attribute]
    if hasattr(arg, "text"):
        return str(arg.text)
    if isinstance(arg, bool):
        return "1" if arg else "0"
    if isinstance(arg, (int, float)):
        return str(arg)
    if isinstance(arg, str):
        escaped = arg.replace("'", "''")
        return f"'{escaped}'"
    return None


# ------------------------------------------------
# JSON helpers retained for engine configuration
# ------------------------------------------------

_ENCODERS: dict[type, Callable[[Any], Any]] = {
    datetime: lambda obj: obj.isoformat(),
    complex: lambda obj: {"real": obj.real, "imag": obj.imag},
}

_DECODERS: dict[type, Callable[[Any], Any]] = {
    datetime: datetime.fromisoformat,
    complex: lambda payload: complex(payload["real"], payload["imag"]),
}


def json_serializer(value: Any) -> str:
    """Convert a value to a JSON string."""

    return json.dumps(value, cls=SqliteJSONEncoder)


def json_deserializer(value: str) -> Any:
    """Convert a JSON string to a Python value."""

    payload = json.loads(value)

    if isinstance(payload, dict) and "__type__" in payload and "__value__" in payload:
        type_name = payload["__type__"]
        converter = next(
            (decoder for t, decoder in _DECODERS.items() if t.__name__ == type_name),
            None,
        )
        if converter is not None:
            return converter(payload["__value__"])
    return payload


class SqliteJSONEncoder(json.JSONEncoder):
    """JSON encoder that handles numpy and datetime types."""

    def default(self, obj: Any) -> Any:  # type: ignore[invalid-method-override]
        import numpy as np

        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, (np.generic, np.number)):
            return obj.item()

        for typ, encoder in _ENCODERS.items():
            if isinstance(obj, typ):
                return {"__type__": typ.__name__, "__value__": encoder(obj)}

        try:
            return super().default(obj)
        except TypeError:
            return f"{obj!s} (unserializable)"


# ------------------------------------------------
# Builders for dataclasses from database rows
# ------------------------------------------------


def _fetch_parameters_for(
    session: Session, simulation_ids: Iterable[int]
) -> dict[int, list[ParameterRecord]]:
    if not simulation_ids:
        return {}
    rows = (
        session.execute(
            select(parameters_table).where(
                parameters_table.c.simulation_id.in_(simulation_ids)
            )
        )
        .mappings()
        .all()
    )
    result: dict[int, list[ParameterRecord]] = {}
    for row in rows:
        parameter = ParameterRecord.from_mapping(row)
        result.setdefault(parameter.simulation_id, []).append(parameter)
    return result


def _fetch_links_for(
    session: Session, simulation_ids: Sequence[int]
) -> dict[int, dict[str, SimulationUID]]:
    if not simulation_ids:
        return {}

    # We join with simulations_table to get the target UID (collection_uid:name)
    stmt = (
        select(
            simulation_links_table.c.source_id,
            simulation_links_table.c.name.label("link_name"),
            simulations_table.c.collection_uid,
            simulations_table.c.name.label("target_name"),
        )
        .select_from(
            simulation_links_table.join(
                simulations_table,
                simulation_links_table.c.target_id == simulations_table.c.id,
            )
        )
        .where(simulation_links_table.c.source_id.in_(simulation_ids))
    )
    rows = session.execute(stmt).mappings().all()

    result: dict[int, dict[str, SimulationUID]] = {}

    for row in rows:
        source_id = row["source_id"]
        link_name = row["link_name"]
        target_uid = SimulationUID(row["collection_uid"], row["target_name"])
        result.setdefault(source_id, {})[link_name] = target_uid
    return result


# ------------------------------------------------
# Upsert statement factories
# ------------------------------------------------


def collections_upsert_stmt(
    data: Sequence[Mapping[str, Any]] | Mapping[str, Any],
) -> Insert:
    """Create an upsert statement for collections.

    Args:
        data: A single dictionary or a sequence of dictionaries representing the
            collections to be inserted or updated.
    """
    payload, _ = _normalize_payload(data, collections_table)
    update_columns = _collect_update_columns(payload, {"uid"})

    stmt = insert(collections_table).values(payload)
    set_clause = {column: getattr(stmt.excluded, column) for column in update_columns}
    return stmt.on_conflict_do_update(
        index_elements=[collections_table.c.uid], set_=set_clause
    )


def simulations_upsert_stmt(
    data: Sequence[Mapping[str, Any]] | Mapping[str, Any],
) -> ReturningInsert[Any]:
    """Create an upsert statement for simulations.

    Args:
        data: A single dictionary or a sequence of dictionaries representing the
            simulations to be inserted or updated.

    Returns:
        A SQLAlchemy Insert statement with a RETURNING clause for the simulation IDs.
    """
    payload, _ = _normalize_payload(data, simulations_table)
    update_columns = _collect_update_columns(payload, {"id"})

    stmt = insert(simulations_table).values(payload)
    set_clause = {column: getattr(stmt.excluded, column) for column in update_columns}
    stmt = stmt.on_conflict_do_update(
        index_elements=[simulations_table.c.collection_uid, simulations_table.c.name],
        set_=set_clause,
    )
    return stmt.returning(simulations_table.c.id)


def parameters_upsert_stmt(
    data: Sequence[Mapping[str, Any]] | Mapping[str, Any],
) -> Insert:
    payload, _ = _normalize_payload(data, parameters_table)
    stmt = insert(parameters_table).values(payload)
    return stmt.on_conflict_do_update(
        index_elements=[parameters_table.c.simulation_id, parameters_table.c.key],
        set_={"value": stmt.excluded.value},
    )


def simulation_links_upsert_stmt(
    data: Sequence[Mapping[str, Any]] | Mapping[str, Any],
) -> Insert:
    payload, _ = _normalize_payload(data, simulation_links_table)
    stmt = insert(simulation_links_table).values(payload)
    return stmt.on_conflict_do_update(
        index_elements=[
            simulation_links_table.c.source_id,
            simulation_links_table.c.name,
        ],
        set_={"target_id": stmt.excluded.target_id},
    )


def delete_collection_stmt(uid: str) -> Delete:
    return delete(collections_table).where(collections_table.c.uid == uid)


def delete_simulation_stmt(collection_uid: str, name: str) -> Delete:
    return delete(simulations_table).where(
        simulations_table.c.collection_uid == collection_uid,
        simulations_table.c.name == name,
    )


def delete_simulation_links_stmt(source_id: int) -> Delete:
    return delete(simulation_links_table).where(
        simulation_links_table.c.source_id == source_id
    )


def _normalize_payload(
    data: Sequence[Mapping[str, Any]] | Mapping[str, Any],
    table: Table,
) -> tuple[Sequence[dict[str, Any]] | dict[str, Any], bool]:
    """Ensure the payload is a list of dictionaries with only valid keys for the table.

    Args:
        data: A single dictionary or a sequence of dictionaries representing the records to be
            inserted or updated.
        table: The SQLAlchemy Table object representing the target database table.

    Returns:
        A tuple containing:
            - A list of dictionaries (or a single dictionary if the input was a single record)
              with keys filtered to match the table's columns.
            - A boolean indicating whether the input was a single record (True) or
              multiple records (False).
    """
    if isinstance(data, Mapping):
        items: Sequence[Mapping[str, Any]] = [data]
        single = True
    else:
        items = list(data)
        single = False

    filtered: list[dict[str, Any]] = []
    valid_keys = set(table.c.keys())
    for record in items:
        filtered.append({key: record[key] for key in record.keys() & valid_keys})
    return (filtered[0] if single else filtered), single


def _collect_update_columns(
    payload: Sequence[dict[str, Any]] | dict[str, Any],
    excluded_keys: set[str],
) -> set[str]:
    """Collect the set of columns to be updated in an upsert operation.

    Args:
        payload: A single dictionary or a sequence of dictionaries representing the
            records to be inserted or updated.
        excluded_keys: A set of keys to exclude from the update operation (e.g., primary keys).
    """
    if isinstance(payload, dict):
        keys = set(payload.keys())
    else:
        keys = set().union(*(record.keys() for record in payload))
    return keys - excluded_keys


# ------------------------------------------------
# Fetch helpers
# ------------------------------------------------


def get_collection_uids(session: Session) -> tuple[str, ...]:
    """Fetch all collection UIDs without loading simulations."""
    rows = session.execute(select(collections_table.c.uid)).scalars()
    return tuple(rows)


def fetch_collection(session: Session, uid: str) -> CollectionRecord | None:
    row = (
        session.execute(select(collections_table).where(collections_table.c.uid == uid))
        .mappings()
        .first()
    )
    if row is None:
        return None

    simulations = fetch_simulations(session, [uid])
    return CollectionRecord.from_mapping(row, simulations)


def fetch_collections(session: Session) -> list[CollectionRecord]:
    rows = session.execute(select(collections_table)).mappings().all()
    return [
        CollectionRecord.from_mapping(row, fetch_simulations(session, [row["uid"]]))
        for row in rows
    ]


def fetch_collection_uid_by_alias(session: Session, alias: str) -> str | None:
    if not alias:
        return None
    alias_key = alias.casefold()
    rows = session.execute(
        select(collections_table.c.uid, collections_table.c.aliases)
    ).all()

    for uid, aliases in rows:
        # if direct match on uid, return it
        if uid.casefold() == alias_key:
            return uid

        if aliases and any(
            str(item).casefold() == alias_key for item in aliases if item
        ):
            return uid
    return None


def fetch_simulation_id(
    session: Session, collection_uid: str, simulation_name: str
) -> None | int:
    """Fetch the ID of a simulation given its collection UID and name."""
    return session.execute(
        select(simulations_table.c.id).where(
            simulations_table.c.collection_uid == collection_uid,
            simulations_table.c.name == simulation_name,
        )
    ).scalar()


def fetch_simulation_uid(session: Session, simulation_id: int) -> SimulationUID | None:
    """Fetch the UID of a simulation given its ID."""
    row = session.execute(
        select(simulations_table.c.collection_uid, simulations_table.c.name).where(
            simulations_table.c.id == simulation_id
        )
    ).first()
    if row is None:
        return None
    return SimulationUID(*row)


def fetch_simulation(
    session: Session, collection_uid: str, name: str
) -> SimulationRecord | None:
    """Fetch a simulation given its collection UID and name."""
    row = (
        session.execute(
            select(simulations_table).where(
                simulations_table.c.collection_uid == collection_uid,
                simulations_table.c.name == name,
            )
        )
        .mappings()
        .first()
    )
    if row is None:
        return None
    parameters_map = _fetch_parameters_for(session, [row["id"]])
    links_map = _fetch_links_for(session, [row["id"]])
    return SimulationRecord.from_mapping(
        row, parameters_map.get(row["id"], []), links_map.get(row["id"], {})
    )


def fetch_simulations_by_uid(
    session: Session, uids: Sequence[tuple[str, str]]
) -> list[SimulationRecord]:
    """Fetch multiple simulations by their collection UID and name."""
    if not uids:
        return []

    stmt = select(simulations_table).where(
        tuple_(simulations_table.c.collection_uid, simulations_table.c.name).in_(uids)
    )
    rows = session.execute(stmt).mappings().all()
    simulation_ids = [row["id"] for row in rows]
    parameters_map = _fetch_parameters_for(session, simulation_ids)
    links_map = _fetch_links_for(session, simulation_ids)
    return [
        SimulationRecord.from_mapping(
            row, parameters_map.get(row["id"], []), links_map.get(row["id"], {})
        )
        for row in rows
    ]


def fetch_simulations(
    session: Session, collection_uids: list[str] | None = None
) -> list[SimulationRecord]:
    """Fetch all simulations in the database."""
    stmt = select(simulations_table)
    if collection_uids is not None:
        stmt = select(simulations_table).where(
            simulations_table.c.collection_uid.in_(collection_uids)
        )

    rows = session.execute(stmt).mappings().all()
    simulation_ids = [row["id"] for row in rows]
    parameters_map = _fetch_parameters_for(session, simulation_ids)
    links_map = _fetch_links_for(session, simulation_ids)
    return [
        SimulationRecord.from_mapping(
            row, parameters_map.get(row["id"], []), links_map.get(row["id"], {})
        )
        for row in rows
    ]


def fetch_parameters(session: Session) -> list[ParameterRecord]:
    rows = session.execute(select(parameters_table)).mappings().all()
    return [ParameterRecord.from_mapping(row) for row in rows]


def fetch_links(session: Session) -> list[LinkRecord]:
    rows = (
        session.execute(
            select(
                simulation_links_table,
                simulations_table.c.collection_uid.label("source_coll_uid"),
                simulations_table.c.name.label("source_name"),
            ).join(
                simulations_table,
                simulation_links_table.c.source_id == simulations_table.c.id,
            )
        )
        .mappings()
        .all()
    )
    return [LinkRecord.from_mapping(row) for row in rows]


def fetch_collection_links(session: Session, collection_uid: str) -> list[LinkRecord]:
    """Fetch all links for a collection."""
    stmt = (
        select(
            simulation_links_table,
            simulations_table.c.collection_uid.label("source_coll_uid"),
            simulations_table.c.name.label("source_name"),
        )
        .join(
            simulations_table,
            simulation_links_table.c.source_id == simulations_table.c.id,
        )
        .where(simulations_table.c.collection_uid == collection_uid)
    )
    rows = session.execute(stmt).mappings().all()
    return [LinkRecord.from_mapping(row) for row in rows]


def fetch_collection_links_map(
    session: Session, collection_uid: str
) -> dict[str, dict[str, str]]:
    """Fetch all links for a collection as a mapping from simulation name to links."""
    sim_src = simulations_table.alias("sim_src")
    sim_target = simulations_table.alias("sim_target")

    stmt = (
        select(
            sim_src.c.name.label("source_name"),
            simulation_links_table.c.name.label("link_name"),
            sim_target.c.collection_uid.label("target_collection_uid"),
            sim_target.c.name.label("target_name"),
        )
        .select_from(
            simulation_links_table.join(
                sim_src,
                simulation_links_table.c.source_id == sim_src.c.id,
            ).join(
                sim_target,
                simulation_links_table.c.target_id == sim_target.c.id,
            )
        )
        .where(sim_src.c.collection_uid == collection_uid)
    )

    rows = session.execute(stmt).mappings().all()
    result: dict[str, dict[str, str]] = {}
    for row in rows:
        source_name = row["source_name"]
        link_name = row["link_name"]
        target_uid = f"{row['target_collection_uid']}{constants.UID_SEPARATOR}{row['target_name']}"
        result.setdefault(source_name, {})[link_name] = target_uid
    return result


def fetch_backlinks(
    session: Session, collection_uid: str, simulation_name: str
) -> Sequence[RowMapping]:
    """Fetch all simulations that link to the given simulation."""
    sim_src = simulations_table.alias("sim_src")
    sim_target = simulations_table.alias("sim_target")

    stmt = (
        select(
            sim_src.c.collection_uid,
            sim_src.c.name.label("source_name"),
            simulation_links_table.c.name.label("link_name"),
        )
        .select_from(
            simulation_links_table.join(
                sim_src,
                simulation_links_table.c.source_id == sim_src.c.id,
            ).join(
                sim_target,
                simulation_links_table.c.target_id == sim_target.c.id,
            )
        )
        .where(
            sim_target.c.collection_uid == collection_uid,
            sim_target.c.name == simulation_name,
        )
    )
    return session.execute(stmt).mappings().all()



"""
This module manages configuration options for bamboost. It supports loading configuration
from a global file (`~/.config/bamboost/config.toml`) and a project configuration in the
standard (`pyproject.toml`).

The configuration is structured using dataclasses, allowing hierarchical and
type-validated configuration handling.

Key Features:
- Detects the root directory of the project based on common anchor files.
- Reads configuration from global and project-specific TOML files.
- Provides structured access to configuration values via dataclasses.
- Supports nested dictionary updates for merging configuration sources.
- Includes an index system for managing paths and database settings.

Attributes:
    ROOT_DIR (Path): The detected root directory of the project.
    config (_Config): The main configuration instance containing paths, options, and index
        settings.

"""

import importlib.util
import sys
from collections.abc import MutableMapping
from dataclasses import dataclass, field, fields
from itertools import chain
from os import makedirs
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Generator,
    Iterable,
    Literal,
    Optional,
    Union,
    get_type_hints,
)

from bamboost._logger import BAMBOOST_LOGGER as log
from bamboost._typing import StrPath
from bamboost.constants import DEFAULT_CONFIG_FILE_NAME, DEFAULT_DATABASE_FILE_NAME
from bamboost.utilities import PathSet

if TYPE_CHECKING:
    from typing_extensions import Self

if sys.version_info < (3, 11):
    import tomli  # type: ignore
else:
    import tomllib as tomli


__all__ = [
    "config",
]

CONFIG_DIR = Path("~/.config/bamboost").expanduser()
CONFIG_FILE = CONFIG_DIR.joinpath(DEFAULT_CONFIG_FILE_NAME)
_LOCAL_DIR = "~/.local/share/bamboost"
LOCAL_DIR = Path(_LOCAL_DIR).expanduser()
CACHE_DIR = Path("~/.cache/bamboost-next").expanduser()
# fmt: off
DEFAULT_EXCLUDE_DIRS: set[str] = {
    # version-control metadata
    ".git", ".hg", ".svn", ".bzr", ".cvs",

    # virtual-envs & package managers
    ".venv", "venv", "env", ".tox", ".nox",
    "node_modules", ".yarn", ".pnp",

    # build / artefact output
    "build", "dist", ".eggs",

    # IDE & editor junk
    ".idea", ".vscode", ".vs",

    # language-specific caches
    "__pycache__", ".mypy_cache", ".pytest_cache", ".ruff_cache",
    ".cache", ".cargo", ".gocache",

    # web / framework caches
    ".next", ".nuxt", ".terser-cache", ".parcel-cache",
    ".vercel", ".serverless", ".aws-sam",

    # infra tools (Terraform etc.)
    ".terraform",

    # vendor copies of dependencies
    "vendor",
}
# fmt: on


def _find_root_dir() -> Optional[Path]:
    """Find the root directory."""

    ANCHORS = [
        ".git",
        "pyproject.toml",
    ]

    cwd = Path.cwd()
    try:
        return next(
            path
            for path in chain([cwd], cwd.parents)
            if any(path.joinpath(anchor).exists() for anchor in ANCHORS)
        )
    except StopIteration:
        log.info("Root directory not found.")
        return None


def _get_global_config(filepath: Path) -> dict[str, Any]:
    """Reads the configuration file and fills the configuration options."""
    try:
        with filepath.open("rb") as f:
            try:
                return tomli.load(f)
            except tomli.TOMLDecodeError as e:
                log.warning(f"Error reading config file: {e}")
                return {}
    except FileNotFoundError:
        log.info("Config file not found or unreadable. Using default settings.")
        return {}


def _get_project_config(project_dir: Path) -> dict[str, Any]:
    """Get the project configuration from bamboost.toml or pyproject.toml."""
    bamboost_path = project_dir.joinpath("bamboost.toml")
    pyproject_path = project_dir.joinpath("pyproject.toml")

    if bamboost_path.is_file():
        try:
            with bamboost_path.open("rb") as f:
                return tomli.load(f)
        except tomli.TOMLDecodeError as e:
            log.warning(f"Error reading bamboost.toml: {e}")
            return {}

    if pyproject_path.is_file():
        try:
            with pyproject_path.open("rb") as f:
                return tomli.load(f).get("tool", {}).get("bamboost", {})
        except tomli.TOMLDecodeError as e:
            log.warning(f"Error reading pyproject.toml: {e}")
            return {}

    log.info("No configuration file found. Using default settings.")
    return {}


def _nested_dict_update(d: MutableMapping, u: MutableMapping) -> MutableMapping:
    for k, v in u.items():
        d[k] = (
            _nested_dict_update(d.get(k, {}), v) if isinstance(v, MutableMapping) else v
        )
    return d


@dataclass
class _Base:
    _field_aliases = {}

    def __repr__(self) -> str:
        s = ""
        max_length = max(len(field.name) for field in fields(self))
        for field in fields(self):
            s += f"{field.name:<{max_length + 1}}: {getattr(self, field.name)}\n"
        return s

    def __getitem__(self, key: str) -> Any:
        """Access the configuration options by key, separated by dots."""
        try:
            current_selection = self
            for attr in key.split("."):
                current_selection = getattr(current_selection, attr)
            return current_selection
        except KeyError:
            raise KeyError(f"Invalid key path: '{key}'")

    def __setitem__(self, key: str, value: Any) -> None:
        """Set the configuration options by key, separated by dots."""
        current_selection = self
        keys = key.split(".")
        for attr in keys[:-1]:
            current_selection = getattr(current_selection, attr)
        setattr(current_selection, keys[-1], value)

    def __getattr__(self, name: str) -> Any:
        # handle aliases
        actual_name = self._field_aliases.get(name, name)
        return super().__getattribute__(actual_name)

    def _ipython_key_completions_(self) -> Generator[str, None, None]:
        for key, obj in self.__dict__.items():
            yield key
            for subkey in getattr(obj, "__dict__", {}).keys():
                yield f"{key}.{subkey}"

    @classmethod
    def from_dict(cls, config: dict, **kwargs) -> "Self":
        """Create an instance of the dataclass from a dictionary of
        configuration values.

        This method performs the following steps:
        1. Identifies valid field names and aliases.
        2. Logs any unknown configuration keys.
        3. Filters the input dictionary to include only valid keys and aliases.
        4. Validates the types of the configuration values.
        5. Creates an instance of the class with the filtered and validated configuration.
        6. Logs any missing fields that are using default values.

        Args:
            config: A dictionary containing configuration key-value pairs.
            **kwargs: Additional keyword arguments to pass to the class constructor

        Returns:
            An instance of the class initialized with the provided configuration.

        Raises:
            No exceptions are raised, but various warnings and errors are logged:
            - Unknown configuration keys are logged as info.
            - Invalid types for configuration values are logged as errors.
            - Type checking errors are logged as info.
            - Missing fields (using defaults) are logged as info.

        Note:
            - Invalid type values are removed from the configuration.
            - Aliases are resolved to their corresponding field names.
        """
        # Get the valid field names
        valid_fields = {f.name for f in fields(cls)}
        aliases = set(cls._field_aliases.keys())

        # Check for unknown keys
        unknown_keys = set(config) - valid_fields - aliases
        for key in unknown_keys:
            log.info(f"Unknown config key: {key}")

        # Filter the input dictionary to only include valid keys
        filtered_config = {k: v for k, v in config.items() if k in valid_fields}
        filtered_config.update(
            {cls._field_aliases[k]: v for k, v in config.items() if k in aliases}
        )

        # Validate type of user and project config values
        resolved_type_hints = get_type_hints(cls)
        for field_def in fields(cls):
            if field_def.name in filtered_config:
                try:
                    if not isinstance(
                        filtered_config[field_def.name],
                        resolved_type_hints[field_def.name],
                    ):
                        log.error(
                            (
                                f"Invalid type for config key '{field_def.name}': {filtered_config[field_def.name]}. "
                                f"Requires {field_def.type} "
                            )
                        )
                        # If the type is invalid, remove the key
                        filtered_config.pop(field_def.name)
                except TypeError:
                    # An error occurred while checking the type, keep the key but log a warning
                    log.info(
                        f"Error checking type for config key '{field_def.name}': {filtered_config[field_def.name]}"
                    )

        # Create the instance
        instance = cls(**filtered_config, **kwargs)

        # Check for missing fields (defaults used)
        for field_def in fields(cls):
            if field_def.name not in filtered_config:
                log.info(
                    f"Config key '{field_def.name}' not set; using default: {getattr(instance, field_def.name)}"
                )

        return instance


# -----------------------------
# Default configuration values
# -----------------------------


@dataclass(repr=False)
class _Paths(_Base):
    """Paths used by bamboost.

    This dataclass contains the paths used by bamboost.

    Attributes:
        localDir: The directory where the local data is stored.
        cacheDir: The directory where the cache is stored.
    """

    localDir: StrPath = field(default=LOCAL_DIR)
    cacheDir: StrPath = field(default=CACHE_DIR)

    def __setattr__(self, name: str, value: Any, /) -> None:
        if isinstance(value, str):
            value = Path(value).expanduser()
        super().__setattr__(name, value)


@dataclass(repr=False)
class _Options(_Base):
    """Core options for bamboost.

    This dataclass contains the core options for bamboost.

    Attributes:
        mpi: If True, the mpi4py package is available and used.
        sortTableKey: The default key to sort the table by.
        sortTableOrder: The default order to sort the table by.
    """

    mpi: bool = field(default=importlib.util.find_spec("mpi4py") is not None)
    sortTableKey: str = field(default="created_at")
    sortTableOrder: str = field(default="desc")

    log_file_lock_severity: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] = (
        field(default="WARNING")
    )
    """The severity level for the log file lock."""

    log_root_only: bool = False
    """If True, only the root logger is used."""

    logLevel: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] = field(
        default="WARNING"
    )
    """The log level for the logger."""

    clipboardCommand: str | None = None
    """The command to use for copying to the clipboard. If None, an error is raised if
    clipboard functionality is used but no command is set."""


@dataclass(repr=False)
class _IndexOptions(_Base):
    """Index options for bamboost.

    This dataclass contains the index options for bamboost.

    Attributes:
        searchPaths: A list of paths to index.
        syncTables: If True, the sqlite tables are synchronized immediatly
            after some queries.
        convertArrays: If True, arrays are converted to np.arrays.
        databaseFileName: The name of the database file.
        databaseFile: The path to the database file.
        isolated: If true, this project manages it's own database. The
            searchPaths are reduced to the project root only.
    """

    _field_aliases = {
        "paths": "searchPaths",
    }

    searchPaths: Iterable[Union[str, Path]] = field(
        default_factory=lambda: PathSet([Path("~").expanduser()])
    )
    """The list of paths to search for collections."""

    excludeDirs: Iterable[str] = field(default_factory=lambda: DEFAULT_EXCLUDE_DIRS)
    """The list of directory names to exclude from the search."""

    extendDefaultExcludeDirs: Iterable[str] | None = None
    """Use this to extend the default exclude directories."""

    syncTables: bool = field(default=True)
    """If True, the sqlite database is updated after some queries to keep it in sync."""

    convertArrays: bool = True
    """If True, sqlite lists are converted to np.arrays. If false, they are left as
    lists."""

    databaseFileName: str = field(default=DEFAULT_DATABASE_FILE_NAME)
    """The basename of the database file."""

    databaseFile: Path = field(init=False)
    """The path to the default database file in the current context. This can be the
    global database or a project-specific database."""

    isolated: bool = False
    """If true, this project manages it's own database. The searchPaths are reduced to the
    project root only."""

    projectDir: Optional[Path] = None
    """The project directory, if found."""

    strictLinks: bool = True
    """If True, an error is raised when trying to set a link and the target is invalid."""

    strictLinksWhenSyncing: bool = False
    """If True, strictLinks is also applied when syncing the database. This can lead to
    errors if files have been moved or deleted since the last sync, but can help to catch
    issues early."""

    def __post_init__(self) -> None:
        # Parse search paths to Path objects
        self.searchPaths = PathSet(
            Path(p).expanduser() if isinstance(p, str) else p for p in self.searchPaths
        )

        # Handle isolated mode
        if self.isolated and self.projectDir:
            self.projectDir.joinpath(".bamboost").mkdir(parents=True, exist_ok=True)
            self.databaseFile = self.projectDir.joinpath(
                ".bamboost", DEFAULT_DATABASE_FILE_NAME
            )
            self.searchPaths = PathSet([self.projectDir])
        else:
            # ensure that localDir exists
            makedirs(LOCAL_DIR, exist_ok=True)
            self.databaseFile = LOCAL_DIR.joinpath(self.databaseFileName)

        # Handle extendDefaultExcludeDirs
        if self.extendDefaultExcludeDirs is not None:
            self.excludeDirs = set((*self.excludeDirs, *self.extendDefaultExcludeDirs))


@dataclass(repr=False, init=False)
class _Config(_Base):
    """Configuration class for bamboost.

    This class manages the configuration options and index settings for bamboost.
    It loads the configuration from a file and provides access to the options
    and index attributes.

    Args:
        project_dir: An optional alternative directory to load the project-based config
            from.

    Attributes:
        paths: Paths used by bamboost.
        options: Configuration options for bamboost.
        index: Index settings for bamboost.
    """

    paths: _Paths
    options: _Options
    index: _IndexOptions

    def __init__(self, project_dir: Optional[StrPath] = None) -> None:
        global_config = _get_global_config(CONFIG_FILE)
        project_dir = Path(project_dir).expanduser() if project_dir else None
        project_dir = project_dir or _find_root_dir()
        if project_dir:
            project_config = _get_project_config(project_dir)
        else:
            project_config = {}

        config = _nested_dict_update(global_config, project_config)

        self.paths = _Paths.from_dict(config.pop("paths", {}))
        self.options = _Options.from_dict(config.pop("options", {}))
        self.index = _IndexOptions.from_dict(
            config.pop("index", {}), projectDir=project_dir
        )

        # Log unknown config options
        self._remainder = config
        for key in config.keys():
            log.info(f"Unknown config table: {key}")

    def __repr__(self) -> str:
        s = str()
        for field in fields(self):
            s += f"> {field.name.upper()}\n"
            s += getattr(self, field.name).__repr__()
            s += "\n"
        return s


# Create the config instance (load the configuration)
config: _Config = _Config()



"""Tests for bamboost._config module."""

from pathlib import Path

from bamboost import _config
from bamboost.constants import DEFAULT_DATABASE_FILE_NAME


class TestIndexOptions:
    """Tests for _IndexOptions configuration."""

    def test_local_dir_created_for_database(self, tmp_path, monkeypatch):
        """Test that local directory is created when loading index config."""
        local_dir = tmp_path / "local-dir"
        monkeypatch.setattr(_config, "LOCAL_DIR", local_dir)

        assert not local_dir.exists()

        index_config = _config._IndexOptions.from_dict({})

        assert local_dir.exists()
        assert index_config.databaseFile == local_dir / DEFAULT_DATABASE_FILE_NAME

    def test_index_options_default_values(self, tmp_path, monkeypatch):
        """Test _IndexOptions default values."""
        local_dir = tmp_path / "local-dir"
        monkeypatch.setattr(_config, "LOCAL_DIR", local_dir)

        index_config = _config._IndexOptions.from_dict({})

        assert isinstance(index_config.databaseFile, Path)
        assert index_config.databaseFile.name == DEFAULT_DATABASE_FILE_NAME


class TestPathsOptions:
    """Tests for _Paths configuration."""

    def test_paths_options_default(self):
        """Test _Paths default initialization."""
        paths = _config._Paths()

        assert isinstance(paths.localDir, Path)
        assert isinstance(paths.cacheDir, Path)

    def test_paths_options_from_dict(self, tmp_path):
        """Test _Paths.from_dict with custom values."""
        custom_local = tmp_path / "custom_local"
        custom_cache = tmp_path / "custom_cache"

        config_dict = {
            "localDir": str(custom_local),
            "cacheDir": str(custom_cache),
        }

        paths = _config._Paths.from_dict(config_dict)

        assert paths.localDir == custom_local
        assert paths.cacheDir == custom_cache


class TestGeneralOptions:
    """Tests for _Options configuration."""

    def test_general_options_default(self):
        """Test _Options default values."""
        options = _config._Options()

        assert options.mpi is False

    def test_general_options_from_dict(self):
        """Test _Options.from_dict."""
        options = _config._Options.from_dict({"mpi": True})

        assert options.mpi is True

    def test_general_options_from_dict_empty(self):
        """Test _Options.from_dict with empty dict."""
        options = _config._Options.from_dict({})

        assert options.mpi is False


class TestConfigClass:
    """Tests for _Config main configuration class."""

    def test_config_initialization(self):
        """Test _Config initialization with defaults."""
        cfg = _config._Config()

        assert isinstance(cfg.paths, _config._Paths)
        assert isinstance(cfg.options, _config._Options)
        assert isinstance(cfg.index, _config._IndexOptions)

    def test_config_nested_update(self):
        """Test that nested dictionary updates work correctly."""
        dict1 = {"a": {"b": 1, "c": 2}}
        dict2 = {"a": {"c": 3, "d": 4}}

        _config._nested_dict_update(dict1, dict2)

        assert dict1 == {"a": {"b": 1, "c": 3, "d": 4}}


class TestConfigUtilities:
    """Tests for configuration utility functions."""

    def test_find_root_dir_not_found(self, tmp_path, monkeypatch):
        """Test _find_root_dir when no anchor files exist."""
        # Change to a temp directory with no anchor files
        monkeypatch.chdir(tmp_path)

        result = _config._find_root_dir()

        # Should return None when no anchor found
        assert result is None

    def test_find_root_dir_with_git(self, tmp_path, monkeypatch):
        """Test _find_root_dir when .git directory exists."""
        git_dir = tmp_path / ".git"
        git_dir.mkdir()

        subdir = tmp_path / "subdir" / "deep"
        subdir.mkdir(parents=True)
        monkeypatch.chdir(subdir)

        result = _config._find_root_dir()

        # Should find the directory containing .git
        assert result == tmp_path

    def test_find_root_dir_with_pyproject(self, tmp_path, monkeypatch):
        """Test _find_root_dir when pyproject.toml exists."""
        pyproject = tmp_path / "pyproject.toml"
        pyproject.touch()

        subdir = tmp_path / "src"
        subdir.mkdir()
        monkeypatch.chdir(subdir)

        result = _config._find_root_dir()

        assert result == tmp_path

    def test_nested_dict_update_simple(self):
        """Test _nested_dict_update with simple dictionaries."""
        target = {"a": 1, "b": 2}
        source = {"b": 3, "c": 4}

        _config._nested_dict_update(target, source)

        assert target == {"a": 1, "b": 3, "c": 4}

    def test_nested_dict_update_deep(self):
        """Test _nested_dict_update with deeply nested dictionaries."""
        target = {"a": {"b": {"c": 1}}}
        source = {"a": {"b": {"d": 2}, "e": 3}}

        _config._nested_dict_update(target, source)

        assert target == {"a": {"b": {"c": 1, "d": 2}, "e": 3}}

    def test_nested_dict_update_mixed_types(self):
        """Test _nested_dict_update with mixed value types."""
        target = {"a": {"b": 1}, "c": [1, 2]}
        source = {"a": {"d": 2}, "c": [3, 4]}

        _config._nested_dict_update(target, source)

        # Non-dict values should be replaced
        assert target == {"a": {"b": 1, "d": 2}, "c": [3, 4]}


class TestGlobalConfig:
    """Tests for global config instance."""

    def test_global_config_exists(self):
        """Test that global config instance exists."""
        assert _config.config is not None
        assert isinstance(_config.config, _config._Config)

    def test_global_config_has_paths(self):
        """Test that global config has paths configuration."""
        assert hasattr(_config.config, "paths")
        assert isinstance(_config.config.paths, _config._Paths)

    def test_global_config_has_options(self):
        """Test that global config has options configuration."""
        assert hasattr(_config.config, "options")
        assert isinstance(_config.config.options, _config._Options)

    def test_global_config_has_index(self):
        """Test that global config has index configuration."""
        assert hasattr(_config.config, "index")
        assert isinstance(_config.config.index, _config._IndexOptions)



from datetime import datetime
from pathlib import Path

import pytest
import yaml

from bamboost.exceptions import InvalidCollectionError
from bamboost.index.base import Index
from bamboost.index.scanner import (
    IDENTIFIER_PREFIX,
    IDENTIFIER_SEPARATOR,
    find_uid_from_path,
    validate_path,
)


@pytest.fixture
def index(tmp_path):
    return Index(sql_file=":memory:", search_paths=[tmp_path])


def id_file_name(uid: str) -> str:
    return f"{IDENTIFIER_PREFIX}{IDENTIFIER_SEPARATOR}{uid}"


def test_find_uid_from_identifier(tmp_path):
    collection_path = tmp_path / "collection2"
    collection_path.mkdir()
    (collection_path / ".bamboost-collection-COLL456").touch()

    uid = find_uid_from_path(collection_path)
    assert uid == "COLL456"


@pytest.mark.parametrize("uid,result", [("123", True), ("345", False)])
def test_validate_path(tmp_path: Path, uid: str, result: bool):
    collection_path = tmp_path / "collection"
    collection_path.mkdir()
    (collection_path / id_file_name("123")).touch()

    assert validate_path(collection_path, uid) == result


def test_insert_collection(index: Index, tmp_path):
    uid = "COLL123"
    collection_path = tmp_path / "collection1"

    index.upsert_collection(uid, collection_path)
    resolved_path = index._get_collection_path(uid)
    assert resolved_path == collection_path


def test_upsert_collection(index: Index, tmp_path):
    uid = "COLL123"
    collection_path = tmp_path / "collection1"
    index.upsert_collection(uid, collection_path)
    assert index._get_collection_path(uid) == collection_path

    # Update the path
    new_collection_path = tmp_path / "collection2"
    index.upsert_collection(uid, new_collection_path)
    assert index._get_collection_path(uid) == new_collection_path
    assert len(index._get_collections()) == 1


def test_resolve_uid(index: Index, tmp_path: Path):
    uid = "COLL789"
    collection_path = tmp_path / "collection3"
    collection_path.mkdir()
    (collection_path / id_file_name(uid)).touch()

    resolved_uid = index.resolve_uid(collection_path)
    assert resolved_uid == uid


def test_resolve_uid_not_found(index: Index, tmp_path: Path):
    collection_path = tmp_path / "collection4"
    collection_path.mkdir()

    with pytest.raises(InvalidCollectionError):
        index.resolve_uid(collection_path)


def test_scan_for_collections(index, tmp_path):
    coll1 = tmp_path / "coll1"
    coll2 = tmp_path / "coll2"
    coll1.mkdir()
    coll2.mkdir()
    (coll1 / id_file_name("UID001")).touch()
    (coll2 / id_file_name("UID002")).touch()

    found = index.scan_for_collections()

    assert len(found) == 2
    assert set(uid for uid, _ in found) == {"UID001", "UID002"}

    stored_uids = {c.uid for c in index.all_collections}
    assert stored_uids == {"UID001", "UID002"}


def test_scan_persists_metadata(index: Index, tmp_path: Path):
    uid = "UIDMETA"
    coll_path = tmp_path / "with_meta"
    coll_path.mkdir()

    metadata_file = coll_path / id_file_name(uid)
    metadata = {
        "uid": uid,
        "created_at": datetime(2024, 5, 1, 12, 30, 0),
        "description": "Sample collection",
        "tags": ["alpha", "beta", "alpha"],
        "aliases": ["demo", "Demo"],
    }
    metadata_file.write_text(yaml.safe_dump(metadata, sort_keys=False))

    index.scan_for_collections()

    collection = index.collection(uid)
    assert collection is not None
    assert collection.description == "Sample collection"
    assert collection.tags == ["alpha", "beta"]
    assert collection.aliases == ["demo"]
    assert collection.created_at == datetime(2024, 5, 1, 12, 30, 0)


def test_resolve_path_with_alias(index: Index, tmp_path: Path):
    uid = "UIDALIAS"
    coll_path = tmp_path / "with_alias"
    coll_path.mkdir()

    metadata_file = coll_path / id_file_name(uid)
    metadata_file.write_text(
        yaml.safe_dump(
            {
                "uid": uid,
                "aliases": ["alias-one"],
            },
            sort_keys=False,
        )
    )

    index.scan_for_collections()

    resolved = index.resolve_path("alias-one")
    assert resolved == coll_path

    fetched = index.collection("ALIAS-ONE")
    assert fetched is not None
    assert fetched.uid == uid



from __future__ import annotations

import uuid
from dataclasses import dataclass
from typing import overload

from typing_extensions import Self

from bamboost import constants
from bamboost.mpi import Communicator


class CollectionUID(str):
    """UID of a collection. If no UID is provided, a new one is generated.

    Note:
        The generated UID is guaranteed to be unique across MPI ranks by broadcasting
        the generated UUID from the root rank.
    """

    def __new__(cls, uid: str | None = None, length: int = 10):
        uid = uid or cls.generate_uid(length)
        return super().__new__(cls, uid.upper())

    @staticmethod
    def generate_uid(length: int) -> str:
        if Communicator._active_comm.rank == 0:
            uid = uuid.uuid4().hex[:length].upper()
        else:
            uid = ""
        uid: str = Communicator._active_comm.bcast(uid, root=0)
        return uid


class SimulationName(str):
    """Name of a simulation. If no name is provided, a new one is generated.

    Args:
        name (Optional[str]): The desired name for the simulation. If not provided,
            a unique name will be generated.
        length (int): The length of the generated name if `name` is not provided.
            Default is 10.

    Note:
        The generated name is guaranteed to be unique across MPI ranks by broadcasting
        the generated UUID from the root rank.
    """

    def __new__(cls, name: str | None = None, length: int = 10):
        name = name or cls.generate_name(length)
        return super().__new__(cls, name)

    @staticmethod
    def generate_name(length: int) -> str:
        if Communicator._active_comm.rank == 0:
            uid = uuid.uuid4().hex[:length]
        else:
            uid = ""
        uid: str = Communicator._active_comm.bcast(uid, root=0)
        return uid


@dataclass(frozen=True, init=False)  # init=False because we handle it in __new__
class SimulationUID:
    """UID of a simulation, consisting of the collection UID and the simulation name.

    Use `str(SimulationUID(...))` to get the string representation of the UID, which is in
    the format `<collection_uid>:<simulation_name>`. The constructor can be called with
    either the string representation or the collection UID and simulation name as separate
    arguments.
    """

    collection_uid: CollectionUID
    simulation_name: SimulationName

    @overload
    def __new__(cls, uid: str | SimulationUID, /) -> Self: ...

    @overload
    def __new__(
        cls,
        collection_uid: CollectionUID | str,
        simulation_name: SimulationName | str,
        /,
    ) -> Self: ...

    def __new__(cls, *args):
        if len(args) == 1 and isinstance(args[0], SimulationUID):
            return args[0]

        instance = super().__new__(cls)

        if len(args) == 1 and isinstance(args[0], str):
            try:
                c_uid, s_name = args[0].split(constants.UID_SEPARATOR, 1)
            except ValueError as exc:
                raise ValueError(
                    f"Invalid SimulationUID string: {args[0]!r}. expected format "
                    f"'<collection_uid>{constants.UID_SEPARATOR}<simulation_name>'"
                ) from exc
            object.__setattr__(instance, "collection_uid", CollectionUID(c_uid))
            object.__setattr__(instance, "simulation_name", SimulationName(s_name))

        elif len(args) == 2:
            object.__setattr__(instance, "collection_uid", CollectionUID(args[0]))
            object.__setattr__(instance, "simulation_name", SimulationName(args[1]))

        else:
            raise ValueError("Invalid arguments for SimulationUID")

        return instance

    def __eq__(self, other):
        if isinstance(other, SimulationUID):
            return (
                self.collection_uid == other.collection_uid
                and self.simulation_name == other.simulation_name
            )
        if isinstance(other, str):
            return str(self) == other
        return NotImplemented

    def __str__(self):
        return f"{self.collection_uid}{constants.UID_SEPARATOR}{self.simulation_name}"



import lazy_loader as _lazy

__getattr__, __dir__, __all__ = _lazy.attach(
    __name__,
    [],
    {
        "file": ["FileMode", "HDF5File", "HDF5Path"],
        "filemap": ["FileMap", "FilteredFileMap"],
    },
)
